"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2022 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class CodegenOptions(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MAX_COUNT_FIELD_NUMBER: builtins.int
    FIXED_COUNT_FIELD_NUMBER: builtins.int
    MAX_SIZE_FIELD_NUMBER: builtins.int
    FIXED_SIZE_FIELD_NUMBER: builtins.int
    USE_CALLBACK_FIELD_NUMBER: builtins.int
    max_count: builtins.int
    """Maximum number of entries for repeated fields."""
    fixed_count: builtins.bool
    """Use a fixed length std::array for repeated fields instead of pw::Vector."""
    max_size: builtins.int
    """Maximum size for bytes and string fields."""
    fixed_size: builtins.bool
    """Use a fixed size std::array for bytes fields instead of pw::Vector."""
    use_callback: builtins.bool
    """Force the use of a callback function for the field."""
    def __init__(
        self,
        *,
        max_count: builtins.int = ...,
        fixed_count: builtins.bool = ...,
        max_size: builtins.int = ...,
        fixed_size: builtins.bool = ...,
        use_callback: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["fixed_count", b"fixed_count", "fixed_size", b"fixed_size", "max_count", b"max_count", "max_size", b"max_size", "use_callback", b"use_callback"]) -> None: ...

global___CodegenOptions = CodegenOptions



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Helpful commands for working with a Git repository."""

import logging
from pathlib import Path
import subprocess
from typing import Collection, Iterable, List, Optional, Pattern, Union

from pw_presubmit.tools import log_run, plural

_LOG = logging.getLogger(__name__)
PathOrStr = Union[Path, str]
PatternOrStr = Union[Pattern, str]

TRACKING_BRANCH_ALIAS = '@{upstream}'
_TRACKING_BRANCH_ALIASES = TRACKING_BRANCH_ALIAS, '@{u}'


def git_stdout(
    *args: PathOrStr, show_stderr=False, repo: PathOrStr = '.'
) -> str:
    return (
        log_run(
            ['git', '-C', str(repo), *args],
            stdout=subprocess.PIPE,
            stderr=None if show_stderr else subprocess.DEVNULL,
            check=True,
        )
        .stdout.decode()
        .strip()
    )


def _ls_files(args: Collection[PathOrStr], repo: Path) -> Iterable[Path]:
    """Returns results of git ls-files as absolute paths."""
    git_root = repo.resolve()
    for file in git_stdout('ls-files', '--', *args, repo=repo).splitlines():
        full_path = git_root / file
        # Modified submodules will show up as directories and should be ignored.
        if full_path.is_file():
            yield full_path


def _diff_names(
    commit: str, pathspecs: Collection[PathOrStr], repo: Path
) -> Iterable[Path]:
    """Returns absolute paths of files changed since the specified commit."""
    git_root = root(repo)
    for file in git_stdout(
        'diff',
        '--name-only',
        '--diff-filter=d',
        commit,
        '--',
        *pathspecs,
        repo=repo,
    ).splitlines():
        full_path = git_root / file
        # Modified submodules will show up as directories and should be ignored.
        if full_path.is_file():
            yield full_path


def tracking_branch(repo_path: Optional[Path] = None) -> Optional[str]:
    """Returns the tracking branch of the current branch.

    Since most callers of this function can safely handle a return value of
    None, suppress exceptions and return None if there is no tracking branch.

    Args:
      repo_path: repo path from which to run commands; defaults to Path.cwd()

    Raises:
      ValueError: if repo_path is not in a Git repository

    Returns:
      the remote tracking branch name or None if there is none
    """
    if repo_path is None:
        repo_path = Path.cwd()

    if not is_repo(repo_path or Path.cwd()):
        raise ValueError(f'{repo_path} is not within a Git repository')

    # This command should only error out if there's no upstream branch set.
    try:
        return git_stdout(
            'rev-parse',
            '--abbrev-ref',
            '--symbolic-full-name',
            TRACKING_BRANCH_ALIAS,
            repo=repo_path,
        )

    except subprocess.CalledProcessError:
        return None


def list_files(
    commit: Optional[str] = None,
    pathspecs: Collection[PathOrStr] = (),
    repo_path: Optional[Path] = None,
) -> List[Path]:
    """Lists files with git ls-files or git diff --name-only.

    Args:
      commit: commit to use as a base for git diff
      pathspecs: Git pathspecs to use in git ls-files or diff
      repo_path: repo path from which to run commands; defaults to Path.cwd()

    Returns:
      A sorted list of absolute paths
    """
    if repo_path is None:
        repo_path = Path.cwd()

    if commit in _TRACKING_BRANCH_ALIASES:
        commit = tracking_branch(repo_path)

    if commit:
        try:
            return sorted(_diff_names(commit, pathspecs, repo_path))
        except subprocess.CalledProcessError:
            _LOG.warning(
                'Error comparing with base revision %s of %s, listing all '
                'files instead of just changed files',
                commit,
                repo_path,
            )

    return sorted(_ls_files(pathspecs, repo_path))


def has_uncommitted_changes(repo: Optional[Path] = None) -> bool:
    """Returns True if the Git repo has uncommitted changes in it.

    This does not check for untracked files.
    """
    if repo is None:
        repo = Path.cwd()

    # Refresh the Git index so that the diff-index command will be accurate.
    # The `git update-index` command isn't reliable when run in parallel with
    # other processes that may touch files in the repo directory, so retry a
    # few times before giving up. The hallmark of this failure mode is the lack
    # of an error message on stderr, so if we see something there we can assume
    # it's some other issue and raise.
    retries = 6
    for i in range(retries):
        try:
            log_run(
                ['git', '-C', repo, 'update-index', '-q', '--refresh'],
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as err:
            if err.stderr or i == retries - 1:
                raise
            continue
    # diff-index exits with 1 if there are uncommitted changes.
    return (
        log_run(
            ['git', '-C', repo, 'diff-index', '--quiet', 'HEAD', '--']
        ).returncode
        == 1
    )


def _describe_constraints(
    git_root: Path,
    repo_path: Path,
    commit: Optional[str],
    pathspecs: Collection[PathOrStr],
    exclude: Collection[Pattern[str]],
) -> Iterable[str]:
    if not git_root.samefile(repo_path):
        yield (
            f'under the {repo_path.resolve().relative_to(git_root.resolve())} '
            'subdirectory'
        )

    if commit in _TRACKING_BRANCH_ALIASES:
        commit = tracking_branch(git_root)
        if commit is None:
            _LOG.warning(
                'Attempted to list files changed since the remote tracking '
                'branch, but the repo is not tracking a branch'
            )

    if commit:
        yield f'that have changed since {commit}'

    if pathspecs:
        paths_str = ', '.join(str(p) for p in pathspecs)
        yield f'that match {plural(pathspecs, "pathspec")} ({paths_str})'

    if exclude:
        yield (
            f'that do not match {plural(exclude, "pattern")} ('
            + ', '.join(p.pattern for p in exclude)
            + ')'
        )


def describe_files(
    git_root: Path,
    repo_path: Path,
    commit: Optional[str],
    pathspecs: Collection[PathOrStr],
    exclude: Collection[Pattern],
    project_root: Optional[Path] = None,
) -> str:
    """Completes 'Doing something to ...' for a set of files in a Git repo."""
    constraints = list(
        _describe_constraints(git_root, repo_path, commit, pathspecs, exclude)
    )

    name = git_root.name
    if project_root and project_root != git_root:
        name = str(git_root.relative_to(project_root))

    if not constraints:
        return f'all files in the {name} repo'

    msg = f'files in the {name} repo'
    if len(constraints) == 1:
        return f'{msg} {constraints[0]}'

    return msg + ''.join(f'\n    - {line}' for line in constraints)


def root(repo_path: PathOrStr = '.', *, show_stderr: bool = True) -> Path:
    """Returns the repository root as an absolute path.

    Raises:
      FileNotFoundError: the path does not exist
      subprocess.CalledProcessError: the path is not in a Git repo
    """
    repo_path = Path(repo_path)
    if not repo_path.exists():
        raise FileNotFoundError(f'{repo_path} does not exist')

    return Path(
        git_stdout(
            'rev-parse',
            '--show-toplevel',
            repo=repo_path if repo_path.is_dir() else repo_path.parent,
            show_stderr=show_stderr,
        )
    )


def within_repo(repo_path: PathOrStr = '.') -> Optional[Path]:
    """Similar to root(repo_path), returns None if the path is not in a repo."""
    try:
        return root(repo_path, show_stderr=False)
    except subprocess.CalledProcessError:
        return None


def is_repo(repo_path: PathOrStr = '.') -> bool:
    """True if the path is tracked by a Git repo."""
    return within_repo(repo_path) is not None


def path(
    repo_path: PathOrStr,
    *additional_repo_paths: PathOrStr,
    repo: PathOrStr = '.',
) -> Path:
    """Returns a path relative to a Git repository's root."""
    return root(repo).joinpath(repo_path, *additional_repo_paths)


def commit_message(commit: str = 'HEAD', repo: PathOrStr = '.') -> str:
    return git_stdout('log', '--format=%B', '-n1', commit, repo=repo)


def commit_author(commit: str = 'HEAD', repo: PathOrStr = '.') -> str:
    return git_stdout('log', '--format=%ae', '-n1', commit, repo=repo)


def commit_hash(
    rev: str = 'HEAD', short: bool = True, repo: PathOrStr = '.'
) -> str:
    """Returns the commit hash of the revision."""
    args = ['rev-parse']
    if short:
        args += ['--short']
    args += [rev]
    return git_stdout(*args, repo=repo)


def discover_submodules(
    superproject_dir: Path, excluded_paths: Collection[PatternOrStr] = ()
) -> List[Path]:
    """Query git and return a list of submodules in the current project.

    Args:
        superproject_dir: Path object to directory under which we are looking
                          for submodules. This will also be included in list
                          returned unless excluded.
        excluded_paths: Pattern or string that match submodules that should not
                        be returned. All matches are done on posix style paths.

    Returns:
        List of "Path"s which were found but not excluded, this includes
        superproject_dir unless excluded.
    """
    discovery_report = git_stdout(
        'submodule',
        'foreach',
        '--quiet',
        '--recursive',
        'echo $toplevel/$sm_path',
        repo=superproject_dir,
    )
    module_dirs = [Path(line) for line in discovery_report.split()]
    # The superproject is omitted in the prior scan.
    module_dirs.append(superproject_dir)

    for exclude in excluded_paths:
        if isinstance(exclude, Pattern):
            for module_dir in reversed(module_dirs):
                if exclude.fullmatch(module_dir.as_posix()):
                    module_dirs.remove(module_dir)
        else:
            for module_dir in reversed(module_dirs):
                if exclude == module_dir.as_posix():
                    module_dirs.remove(module_dir)

    return module_dirs



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/echo.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_rpc/echo.proto',
  package='pw.rpc',
  syntax='proto3',
  serialized_options=b'\n\030dev.pigweed.pw_rpc.proto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x11pw_rpc/echo.proto\x12\x06pw.rpc\"\x1a\n\x0b\x45\x63hoMessage\x12\x0b\n\x03msg\x18\x01 \x01(\t2A\n\x0b\x45\x63hoService\x12\x32\n\x04\x45\x63ho\x12\x13.pw.rpc.EchoMessage\x1a\x13.pw.rpc.EchoMessage\"\x00\x42\x1a\n\x18\x64\x65v.pigweed.pw_rpc.protob\x06proto3'
)




_ECHOMESSAGE = _descriptor.Descriptor(
  name='EchoMessage',
  full_name='pw.rpc.EchoMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='msg', full_name='pw.rpc.EchoMessage.msg', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29,
  serialized_end=55,
)

DESCRIPTOR.message_types_by_name['EchoMessage'] = _ECHOMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EchoMessage = _reflection.GeneratedProtocolMessageType('EchoMessage', (_message.Message,), {
  'DESCRIPTOR' : _ECHOMESSAGE,
  '__module__' : 'pw_rpc.echo_pb2'
  # @@protoc_insertion_point(class_scope:pw.rpc.EchoMessage)
  })
_sym_db.RegisterMessage(EchoMessage)


DESCRIPTOR._options = None

_ECHOSERVICE = _descriptor.ServiceDescriptor(
  name='EchoService',
  full_name='pw.rpc.EchoService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=57,
  serialized_end=122,
  methods=[
  _descriptor.MethodDescriptor(
    name='Echo',
    full_name='pw.rpc.EchoService.Echo',
    index=0,
    containing_service=None,
    input_type=_ECHOMESSAGE,
    output_type=_ECHOMESSAGE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_ECHOSERVICE)

DESCRIPTOR.services_by_name['EchoService'] = _ECHOSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""The env module defines the environment variables used by Pigweed."""

from pathlib import Path
from typing import Optional

from pw_cli import envparse


def pigweed_environment_parser() -> envparse.EnvironmentParser:
    """Defines Pigweed's environment variables on an EnvironmentParser."""
    parser = envparse.EnvironmentParser(prefix='PW_')

    parser.add_var('PW_BOOTSTRAP_PYTHON')
    parser.add_var('PW_ENABLE_PRESUBMIT_HOOK_WARNING', default=False)
    parser.add_var('PW_EMOJI', type=envparse.strict_bool, default=False)
    parser.add_var('PW_ENVSETUP')
    parser.add_var('PW_ENVSETUP_FULL')
    parser.add_var(
        'PW_ENVSETUP_NO_BANNER', type=envparse.strict_bool, default=False
    )
    parser.add_var(
        'PW_ENVSETUP_QUIET', type=envparse.strict_bool, default=False
    )
    parser.add_var('PW_ENVIRONMENT_ROOT', type=Path)
    parser.add_var('PW_PACKAGE_ROOT', type=Path)
    parser.add_var('PW_PROJECT_ROOT', type=Path)
    parser.add_var('PW_ROOT', type=Path)
    parser.add_var(
        'PW_DISABLE_ROOT_GIT_REPO_CHECK',
        type=envparse.strict_bool,
        default=False,
    )
    parser.add_var('PW_SKIP_BOOTSTRAP')
    parser.add_var('PW_SUBPROCESS', type=envparse.strict_bool, default=False)
    parser.add_var('PW_USE_COLOR', type=envparse.strict_bool, default=False)
    parser.add_var('PW_USE_GCS_ENVSETUP', type=envparse.strict_bool)

    parser.add_allowed_suffix('_CIPD_INSTALL_DIR')

    parser.add_var(
        'PW_ENVSETUP_DISABLE_SPINNER', type=envparse.strict_bool, default=False
    )
    parser.add_var('PW_DOCTOR_SKIP_CIPD_CHECKS')
    parser.add_var(
        'PW_ACTIVATE_SKIP_CHECKS', type=envparse.strict_bool, default=False
    )

    parser.add_var('PW_BANNER_FUNC')
    parser.add_var('PW_BRANDING_BANNER')
    parser.add_var('PW_BRANDING_BANNER_COLOR', default='magenta')

    parser.add_var(
        'PW_PRESUBMIT_DISABLE_SUBPROCESS_CAPTURE', type=envparse.strict_bool
    )

    parser.add_var('PW_CONSOLE_CONFIG_FILE')
    parser.add_var('PW_ENVIRONMENT_NO_ERROR_ON_UNRECOGNIZED')

    parser.add_var('PW_CIPD_SERVICE_ACCOUNT_JSON')

    # RBE environment variables
    parser.add_var('PW_USE_RBE', default=False)
    parser.add_var('PW_RBE_DEBUG', default=False)
    parser.add_var('PW_RBE_CLANG_CONFIG', default='')
    parser.add_var('PW_RBE_ARM_GCC_CONFIG', default='')

    return parser


# Internal: memoize environment parsing to avoid unnecessary computation in
# multiple calls to pigweed_environment().
_memoized_environment: Optional[envparse.EnvNamespace] = None


def pigweed_environment() -> envparse.EnvNamespace:
    """Returns Pigweed's parsed environment."""
    global _memoized_environment  # pylint: disable=global-statement

    if _memoized_environment is None:
        _memoized_environment = pigweed_environment_parser().parse_env()

    return _memoized_environment



#!/usr/bin/env python
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Installs and then runs cipd.

This script installs cipd in ./tools/ (if necessary) and then executes it,
passing through all arguments.

Must be tested with Python 2 and Python 3.
"""

from __future__ import print_function

import hashlib
import os
import platform
import ssl
import subprocess
import sys
import base64

try:
    import httplib  # type: ignore
except ImportError:
    import http.client as httplib  # type: ignore[no-redef]

try:
    import urlparse  # type: ignore
except ImportError:
    import urllib.parse as urlparse  # type: ignore[no-redef]

# Generated from the following command. May need to be periodically rerun.
# $ cipd ls infra/tools/cipd | perl -pe "s[.*/][];s/^/    '/;s/\s*$/',\n/;"
SUPPORTED_PLATFORMS = (
    'aix-ppc64',
    'linux-386',
    'linux-amd64',
    'linux-arm64',
    'linux-armv6l',
    'linux-mips64',
    'linux-mips64le',
    'linux-mipsle',
    'linux-ppc64',
    'linux-ppc64le',
    'linux-s390x',
    'mac-amd64',
    'mac-arm64',
    'windows-386',
    'windows-amd64',
)


class UnsupportedPlatform(Exception):
    pass


try:
    SCRIPT_DIR = os.path.dirname(__file__)
except NameError:  # __file__ not defined.
    try:
        SCRIPT_DIR = os.path.join(
            os.environ['PW_ROOT'],
            'pw_env_setup',
            'py',
            'pw_env_setup',
            'cipd_setup',
        )
    except KeyError:
        raise Exception('Environment variable PW_ROOT not set')

VERSION_FILE = os.path.join(SCRIPT_DIR, '.cipd_version')
DIGESTS_FILE = VERSION_FILE + '.digests'

# Put CIPD client in tools so that users can easily get it in their PATH.
CIPD_HOST = 'chrome-infra-packages.appspot.com'

try:
    PW_ROOT = os.environ['PW_ROOT']
except KeyError:
    try:
        with open(os.devnull, 'w') as outs:
            PW_ROOT = (
                subprocess.check_output(
                    ['git', 'rev-parse', '--show-toplevel'],
                    stderr=outs,
                )
                .strip()
                .decode('utf-8')
            )
    except subprocess.CalledProcessError:
        PW_ROOT = ''

# Get default install dir from environment since args cannot always be passed
# through this script (args are passed as-is to cipd).
if 'CIPD_PY_INSTALL_DIR' in os.environ:
    DEFAULT_INSTALL_DIR = os.environ['CIPD_PY_INSTALL_DIR']
elif PW_ROOT:
    DEFAULT_INSTALL_DIR = os.path.join(PW_ROOT, '.cipd')
else:
    DEFAULT_INSTALL_DIR = ''


def platform_normalized():
    """Normalize platform into format expected in CIPD paths."""

    try:
        os_name = platform.system().lower()
        return {
            'linux': 'linux',
            'mac': 'mac',
            'darwin': 'mac',
            'windows': 'windows',
        }[os_name]
    except KeyError:
        raise Exception('unrecognized os: {}'.format(os_name))


def arch_normalized(rosetta=False):
    """Normalize arch into format expected in CIPD paths."""

    machine = platform.machine()
    if platform_normalized() == 'mac' and rosetta:
        return 'amd64'
    if machine.startswith(('arm', 'aarch')):
        machine = machine.replace('aarch', 'arm')
        if machine == 'arm64':
            return machine
        return 'armv6l'
    if machine.endswith('64'):
        return 'amd64'
    if machine.endswith('86'):
        return '386'
    raise Exception('unrecognized arch: {}'.format(machine))


def platform_arch_normalized(rosetta=False):
    return '{}-{}'.format(platform_normalized(), arch_normalized(rosetta))


def user_agent():
    """Generate a user-agent based on the project name and current hash."""

    try:
        rev = subprocess.check_output(
            ['git', '-C', SCRIPT_DIR, 'rev-parse', 'HEAD']
        ).strip()
    except subprocess.CalledProcessError:
        rev = '???'

    if isinstance(rev, bytes):
        rev = rev.decode()

    return 'pigweed-infra/tools/{}'.format(rev)


def actual_hash(path):
    """Hash the file at path and return it."""

    hasher = hashlib.sha256()
    with open(path, 'rb') as ins:
        hasher.update(ins.read())
    return hasher.hexdigest()


def expected_hash(rosetta=False):
    """Pulls expected hash from digests file."""

    expected_plat = platform_arch_normalized(rosetta)

    with open(DIGESTS_FILE, 'r') as ins:
        for line in ins:
            line = line.strip()
            if line.startswith('#') or not line:
                continue
            plat, hashtype, hashval = line.split()
            if hashtype == 'sha256' and plat == expected_plat:
                return hashval
    raise Exception('platform {} not in {}'.format(expected_plat, DIGESTS_FILE))


def https_connect_with_proxy(target_url):
    """Create HTTPSConnection with proxy support."""

    proxy_env = os.environ.get('HTTPS_PROXY') or os.environ.get('https_proxy')
    if proxy_env in (None, ''):
        conn = httplib.HTTPSConnection(target_url)
        return conn

    url = urlparse.urlparse(proxy_env)
    conn = httplib.HTTPSConnection(url.hostname, url.port)
    headers = {}
    if url.username and url.password:
        auth = '%s:%s' % (url.username, url.password)
        py_version = sys.version_info.major
        if py_version >= 3:
            headers['Proxy-Authorization'] = 'Basic ' + str(
                base64.b64encode(auth.encode()).decode()
            )
        else:
            headers['Proxy-Authorization'] = 'Basic ' + base64.b64encode(auth)
    conn.set_tunnel(target_url, 443, headers)
    return conn


def client_bytes(rosetta=False):
    """Pull down the CIPD client and return it as a bytes object.

    Often CIPD_HOST returns a 302 FOUND with a pointer to
    storage.googleapis.com, so this needs to handle redirects, but it
    shouldn't require the initial response to be a redirect either.
    """

    with open(VERSION_FILE, 'r') as ins:
        version = ins.read().strip()

    try:
        conn = https_connect_with_proxy(CIPD_HOST)
    except AttributeError:
        print('=' * 70)
        print(
            '''
It looks like this version of Python does not support SSL. This is common
when using Homebrew. If using Homebrew please run the following commands.
If not using Homebrew check how your version of Python was built.

brew install openssl  # Probably already installed, but good to confirm.
brew uninstall python && brew install python
'''.strip()
        )
        print('=' * 70)
        raise

    full_platform = platform_arch_normalized(rosetta)
    if full_platform not in SUPPORTED_PLATFORMS:
        raise UnsupportedPlatform(full_platform)

    path = '/client?platform={}&version={}'.format(full_platform, version)

    for _ in range(10):
        try:
            conn.request('GET', path)
            res = conn.getresponse()
            # Have to read the response before making a new request, so make
            # sure we always read it.
            content = res.read()
        except ssl.SSLError:
            print(
                '\n'
                'Bootstrap: SSL error in Python when downloading CIPD client.\n'
                'If using system Python try\n'
                '\n'
                '    sudo pip install certifi\n'
                '\n'
                'And if on the system Python on a Mac try\n'
                '\n'
                '    /Applications/Python 3.6/Install Certificates.command\n'
                '\n'
                'If using Homebrew Python try\n'
                '\n'
                '    brew install openssl\n'
                '    brew uninstall python\n'
                '    brew install python\n'
                '\n'
                "If those don't work, address all the potential issues shown \n"
                'by the following command.\n'
                '\n'
                '    brew doctor\n'
                '\n'
                "Otherwise, check that your machine's Python can use SSL, "
                'testing with the httplib module on Python 2 or http.client on '
                'Python 3.',
                file=sys.stderr,
            )
            raise

        # Found client bytes.
        if res.status == httplib.OK:  # pylint: disable=no-else-return
            return content

        # Redirecting to another location.
        elif res.status == httplib.FOUND:
            location = res.getheader('location')
            url = urlparse.urlparse(location)
            if url.netloc != conn.host:
                conn = https_connect_with_proxy(url.netloc)
            path = '{}?{}'.format(url.path, url.query)

        # Some kind of error in this response.
        else:
            break

    raise Exception(
        'failed to download client from https://{}{}'.format(CIPD_HOST, path)
    )


def bootstrap(
    client,
    silent=('PW_ENVSETUP_QUIET' in os.environ),
    rosetta=False,
):
    """Bootstrap cipd client installation."""

    client_dir = os.path.dirname(client)
    if not os.path.isdir(client_dir):
        os.makedirs(client_dir)

    if not silent:
        print(
            'Bootstrapping cipd client for {}'.format(
                platform_arch_normalized(rosetta)
            )
        )

    tmp_path = client + '.tmp'
    with open(tmp_path, 'wb') as tmp:
        tmp.write(client_bytes(rosetta))

    expected = expected_hash(rosetta=rosetta)
    actual = actual_hash(tmp_path)

    if expected != actual:
        raise Exception(
            'digest of downloaded CIPD client is incorrect, '
            'check that digests file is current'
        )

    os.chmod(tmp_path, 0o755)
    os.rename(tmp_path, client)


def selfupdate(client):
    """Update cipd client."""

    cmd = [
        client,
        'selfupdate',
        '-version-file',
        VERSION_FILE,
        '-service-url',
        'https://{}'.format(CIPD_HOST),
    ]
    subprocess.check_call(cmd)


def _default_client(install_dir):
    client = os.path.join(install_dir, 'cipd')
    if os.name == 'nt':
        client += '.exe'
    return client


def init(
    install_dir=DEFAULT_INSTALL_DIR,
    silent=False,
    client=None,
    rosetta=False,
):
    """Install/update cipd client."""

    if not client:
        client = _default_client(install_dir)

    os.environ['CIPD_HTTP_USER_AGENT_PREFIX'] = user_agent()

    if not os.path.isfile(client):
        bootstrap(client, silent, rosetta=rosetta)

    try:
        selfupdate(client)
    except subprocess.CalledProcessError:
        print(
            'CIPD selfupdate failed. Bootstrapping then retrying...',
            file=sys.stderr,
        )
        bootstrap(client, rosetta=rosetta)
        selfupdate(client)

    return client


def main(install_dir=DEFAULT_INSTALL_DIR, silent=False):
    """Install/update cipd client."""

    client = _default_client(install_dir)

    try:
        init(install_dir=install_dir, silent=silent, client=client)

    except UnsupportedPlatform:
        # Don't show help message below for this exception.
        raise

    except Exception:
        print(
            'Failed to initialize CIPD. Run '
            '`CIPD_HTTP_USER_AGENT_PREFIX={user_agent}/manual {client} '
            "selfupdate -version-file '{version_file}'` "
            'to diagnose if this is persistent.'.format(
                user_agent=user_agent(),
                client=client,
                version_file=VERSION_FILE,
            ),
            file=sys.stderr,
        )
        raise

    return client


if __name__ == '__main__':
    client_exe = main()
    subprocess.check_call([client_exe] + sys.argv[1:])



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Library to analyze timestamp."""

from typing import List
import datetime
from pw_chrono_protos import chrono_pb2

_UTC_EPOCH = datetime.datetime(1970, 1, 1, 00, 00, 00)

_UNKNOWN = chrono_pb2.EpochType.Enum.UNKNOWN
_TIME_SINCE_BOOT = chrono_pb2.EpochType.Enum.TIME_SINCE_BOOT
_UTC_WALL_CLOCK = chrono_pb2.EpochType.Enum.UTC_WALL_CLOCK


def process_snapshot(serialized_snapshot: bytes):
    captured_timestamps = chrono_pb2.SnapshotTimestamps()
    captured_timestamps.ParseFromString(serialized_snapshot)
    return timestamp_output(captured_timestamps)


def timestamp_output(timestamps: chrono_pb2.SnapshotTimestamps):
    output: List[str] = []
    if not timestamps.timestamps:
        return ''

    plural = '' if len(timestamps.timestamps) == 1 else 's'
    output.append(f'Snapshot capture timestamp{plural}')
    for timepoint in timestamps.timestamps:
        time = timestamp_snapshot_analyzer(timepoint)
        clock_epoch_type = timepoint.clock_parameters.epoch_type
        if clock_epoch_type == _TIME_SINCE_BOOT:
            output.append(f'  Time since boot:   {time}')
        elif clock_epoch_type == _UTC_WALL_CLOCK:
            utc_time = time + _UTC_EPOCH
            output.append(f'  UTC time:   {utc_time}')
        else:
            output.append(f'  Time since unknown epoch {_UNKNOWN}:   unknown')

    return '\n'.join(output)


def timestamp_snapshot_analyzer(
    captured_timepoint: chrono_pb2.TimePoint,
) -> datetime.timedelta:
    ticks = captured_timepoint.timestamp
    clock_period = (
        captured_timepoint.clock_parameters.tick_period_seconds_numerator
        / captured_timepoint.clock_parameters.tick_period_seconds_denominator
    )
    elapsed_seconds = ticks * clock_period

    time_delta = datetime.timedelta(seconds=elapsed_seconds)

    return time_delta



# Copyright 2019 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Copies built host tools into Pigweed's host_tools directory."""

import argparse
import logging
from pathlib import Path
import shutil
import sys
from typing import Optional

import pw_cli.log

_LOG = logging.getLogger(__name__)


def argument_parser(
    parser: Optional[argparse.ArgumentParser] = None,
) -> argparse.ArgumentParser:
    """Registers the script's arguments on an argument parser."""

    if parser is None:
        parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        '--dst', type=Path, required=True, help='Path to host tools directory'
    )
    parser.add_argument('--name', help='Name for the installed tool')
    parser.add_argument(
        '--out-root',
        type=Path,
        required=True,
        help='Root of Ninja out directory',
    )
    parser.add_argument(
        '--src', type=Path, required=True, help='Path to host tool executable'
    )

    return parser


def main() -> int:
    """Copies a host tool into a destination directory."""
    args = argument_parser().parse_args()

    if not args.src.is_file():
        _LOG.error('%s is not a file', args.src)
        return 1

    args.dst.mkdir(parents=True, exist_ok=True)

    if args.name is not None:
        if '/' in args.name:
            _LOG.error('Host tool name cannot contain "/"')
            return 1
        name = args.name
    else:
        name = args.src.name

    try:
        shutil.copy2(args.src, args.dst.joinpath(name))
    except OSError as err:
        _LOG.error('%s', err)

        # Errno 26 (text file busy) indicates that a host tool binary is
        # currently running.
        # TODO(frolv): Check if this works on Windows.
        if err.errno == 26:
            _LOG.error('')
            _LOG.error('  %s has been rebuilt but cannot be', name)
            _LOG.error('  copied into the host tools directory:')
            _LOG.error('')
            _LOG.error(
                '    %s', args.dst.relative_to(args.out_root).joinpath(name)
            )
            _LOG.error('')
            _LOG.error('  This can occur if the program is already running.')
            _LOG.error(
                '  If it is running, exit it and try re-running the build.'
            )
            _LOG.error('')

        return 1

    return 0


if __name__ == '__main__':
    pw_cli.log.install()
    sys.exit(main())



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# Inspired by
# https://fuchsia.googlesource.com/infra/recipes/+/336933647862a1a9718b4ca18f0a67e89c2419f8/recipe_modules/ninja/resources/ninja_wrapper.py
"""Extracts a concise error from a ninja log."""

from pathlib import Path
import re

_RULE_RE = re.compile(r'^\s*\[\d+/\d+\] (\S+)')
_FAILED_RE = re.compile(r'^\s*FAILED: (.*)$')
_FAILED_END_RE = re.compile(r'^\s*ninja: build stopped:.*')


def parse_ninja_stdout(ninja_stdout: Path) -> str:
    """Extract an error summary from ninja output."""

    failure_begins = False
    failure_lines = []
    last_line = ''

    with ninja_stdout.open() as ins:
        for line in ins:
            # Trailing whitespace isn't significant, as it doesn't affect the
            # way the line shows up in the logs. However, leading whitespace may
            # be significant, especially for compiler error messages.
            line = line.rstrip()
            if failure_begins:
                if not _RULE_RE.match(line) and not _FAILED_END_RE.match(line):
                    failure_lines.append(line)
                else:
                    # Output of failed step ends, save its info.
                    failure_begins = False
            else:
                failed_nodes_match = _FAILED_RE.match(line)
                failure_begins = False
                if failed_nodes_match:
                    failure_begins = True
                    failure_lines.extend([last_line, line])
            last_line = line

    # Remove "Requirement already satisfied:" lines since many of those might
    # be printed during Python installation, and they usually have no relevance
    # to the actual error.
    failure_lines = [
        x
        for x in failure_lines
        if not x.lstrip().startswith('Requirement already satisfied:')
    ]

    result = '\n'.join(failure_lines)
    return re.sub(r'\n+', '\n', result)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Metadata signing facilities."""

import argparse
from pathlib import Path

from pw_software_update import keys
from pw_software_update.tuf_pb2 import SignedRootMetadata
from pw_software_update.update_bundle_pb2 import UpdateBundle


def sign_root_metadata(
    root_metadata: SignedRootMetadata, root_key_pem: bytes
) -> SignedRootMetadata:
    """Signs or re-signs a Root Metadata.

    Args:
      root_metadata: A SignedRootMetadata to be signed/re-signed.
      root_key_pem: The Root signing key in PEM.
    """

    signature = keys.create_ecdsa_signature(
        root_metadata.serialized_root_metadata, root_key_pem
    )
    root_metadata.signatures.append(signature)

    return root_metadata


def sign_update_bundle(
    bundle: UpdateBundle, targets_key_pem: bytes
) -> UpdateBundle:
    """Signs or re-signs an update bundle.

    Args:
      bundle: An UpdateBundle to be signed/re-signed.
      targets_key_pem: The targets signing key in PEM.
    """
    bundle.targets_metadata['targets'].signatures.append(
        keys.create_ecdsa_signature(
            bundle.targets_metadata['targets'].serialized_targets_metadata,
            targets_key_pem,
        )
    )
    return bundle


def parse_args():
    """Parse CLI arguments."""
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        '--root-metadata',
        type=Path,
        required=False,
        help='Path to the root metadata to be signed',
    )

    parser.add_argument(
        '--bundle',
        type=Path,
        required=False,
        help='Path to the bundle to be signed',
    )

    parser.add_argument(
        '--output',
        type=Path,
        required=False,
        help=(
            'Path to save the signed root metadata or bundle '
            'to; Defaults to the input path if unspecified'
        ),
    )

    parser.add_argument(
        '--key', type=Path, required=True, help='Path to the signing key'
    )

    args = parser.parse_args()

    if not (args.root_metadata or args.bundle):
        parser.error('either "--root-metadata" or "--bundle" must be specified')
    if args.root_metadata and args.bundle:
        parser.error('"--root-metadata" and "--bundle" are mutually exclusive')

    return args


def main(root_metadata: Path, bundle: Path, key: Path, output: Path) -> None:
    """Signs or re-signs a root metadata or an update bundle."""
    if root_metadata:
        signed_root_metadata = sign_root_metadata(
            SignedRootMetadata.FromString(root_metadata.read_bytes()),
            key.read_bytes(),
        )

        if not output:
            output = root_metadata
        output.write_bytes(signed_root_metadata.SerializeToString())
    else:
        signed_bundle = sign_update_bundle(
            UpdateBundle.FromString(bundle.read_bytes()), key.read_bytes()
        )

        if not output:
            output = bundle
        output.write_bytes(signed_bundle.SerializeToString())


if __name__ == '__main__':
    main(**vars(parse_args()))



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2021 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import pw_software_update.tuf_pb2
import sys
import typing

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class UpdateBundle(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class TargetsMetadataEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> pw_software_update.tuf_pb2.SignedTargetsMetadata: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: pw_software_update.tuf_pb2.SignedTargetsMetadata | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    class TargetPayloadsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.bytes
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.bytes = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    TIMESTAMP_METADATA_FIELD_NUMBER: builtins.int
    SNAPSHOT_METADATA_FIELD_NUMBER: builtins.int
    TARGETS_METADATA_FIELD_NUMBER: builtins.int
    TARGET_PAYLOADS_FIELD_NUMBER: builtins.int
    ROOT_METADATA_FIELD_NUMBER: builtins.int
    @property
    def timestamp_metadata(self) -> pw_software_update.tuf_pb2.SignedTimestampMetadata:
        """The timestamp role is used for freshness check of the snapshot. Any
        project-specific update metadata should go in the top-level
        targets_metadata or with the TargetFile information
        """
    @property
    def snapshot_metadata(self) -> pw_software_update.tuf_pb2.SignedSnapshotMetadata:
        """The snapshot role is used to ensure that the collection of targets_metadata
        files is securely consistent (no target metadata mix and match). Any
        project-specific update metadata should go in the top-level
        targets_metadata or with the TargetFile information
        """
    @property
    def targets_metadata(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, pw_software_update.tuf_pb2.SignedTargetsMetadata]:
        """Map of target metadata name to target metadata.
        Target metadata name can be an arbitrary name or a path that describes
        where the file lives relative to the base directory of the repository, as
        described in the snapshot metadata. e.g. "path/to/target/0".
        """
    @property
    def target_payloads(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.bytes]:
        """Map of target file name to target payload bytes.
        Target file name can be an arbitrary name or a path that describes where
        the file lives relative to the base directory of the repository, as
        described in the target metadata. e.g. "path/to/amber_tools/0".
        """
    @property
    def root_metadata(self) -> pw_software_update.tuf_pb2.SignedRootMetadata:
        """If present, a client will attempt to upgrade its on-device trusted root
        metadata to the root metadata included in the bundle, following the
        standard "Update the root role" flow specified in the TUF spec, but
        without "version climbing".

        The exact steps are:
        1. Check if there is a root metadata in the bundle.
        2. If the root metadata IS NOT included, assume on-device root metadata
           is up-to-date and continue with the rest of metadata verification.
        3. If the root metadata IS included, verify the new root metadata using
           the on-device root metadata.
        4. If the verification is successful, persist new root metadata and
           continue with the rest of metadata verification. Otherwise abort the
           update session.

        The key deviation from standard flow is the client assumes it can always
        directly upgrade to the single new root metadata in the update bundle,
        without any step-stone history root metadata. This works only because
        we are not supporting (more than 1) root key rotations.
        """
    def __init__(
        self,
        *,
        timestamp_metadata: pw_software_update.tuf_pb2.SignedTimestampMetadata | None = ...,
        snapshot_metadata: pw_software_update.tuf_pb2.SignedSnapshotMetadata | None = ...,
        targets_metadata: collections.abc.Mapping[builtins.str, pw_software_update.tuf_pb2.SignedTargetsMetadata] | None = ...,
        target_payloads: collections.abc.Mapping[builtins.str, builtins.bytes] | None = ...,
        root_metadata: pw_software_update.tuf_pb2.SignedRootMetadata | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_root_metadata", b"_root_metadata", "_snapshot_metadata", b"_snapshot_metadata", "_timestamp_metadata", b"_timestamp_metadata", "root_metadata", b"root_metadata", "snapshot_metadata", b"snapshot_metadata", "timestamp_metadata", b"timestamp_metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_root_metadata", b"_root_metadata", "_snapshot_metadata", b"_snapshot_metadata", "_timestamp_metadata", b"_timestamp_metadata", "root_metadata", b"root_metadata", "snapshot_metadata", b"snapshot_metadata", "target_payloads", b"target_payloads", "targets_metadata", b"targets_metadata", "timestamp_metadata", b"timestamp_metadata"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_root_metadata", b"_root_metadata"]) -> typing_extensions.Literal["root_metadata"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_snapshot_metadata", b"_snapshot_metadata"]) -> typing_extensions.Literal["snapshot_metadata"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_timestamp_metadata", b"_timestamp_metadata"]) -> typing_extensions.Literal["timestamp_metadata"] | None: ...

global___UpdateBundle = UpdateBundle

class Manifest(google.protobuf.message.Message):
    """Update bundle metadata
    Designed to inform the update server what the device currently has in-place.
    Also used to persist the TUF metadata for use in the verification process.
    Stored manifest is only written/erased by the update service. In all other
    contexts the stored manifest is considered read-only.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class TargetsMetadataEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> pw_software_update.tuf_pb2.TargetsMetadata: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: pw_software_update.tuf_pb2.TargetsMetadata | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    TARGETS_METADATA_FIELD_NUMBER: builtins.int
    USER_MANIFEST_FIELD_NUMBER: builtins.int
    @property
    def targets_metadata(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, pw_software_update.tuf_pb2.TargetsMetadata]: ...
    user_manifest: builtins.bytes
    """Insert user manifest target file content here"""
    def __init__(
        self,
        *,
        targets_metadata: collections.abc.Mapping[builtins.str, pw_software_update.tuf_pb2.TargetsMetadata] | None = ...,
        user_manifest: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_user_manifest", b"_user_manifest", "user_manifest", b"user_manifest"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_user_manifest", b"_user_manifest", "targets_metadata", b"targets_metadata", "user_manifest", b"user_manifest"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_user_manifest", b"_user_manifest"]) -> typing_extensions.Literal["user_manifest"] | None: ...

global___Manifest = Manifest



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Module for low-level HDLC protocol features."""

from typing import Tuple

import zlib

# Special flag character for delimiting HDLC frames.
FLAG = 0x7E

# Special character for escaping other special characters in a frame.
ESCAPE = 0x7D

# Characters allowed after a 0x7d escape character.
VALID_ESCAPED_BYTES = 0x5D, 0x5E

# Maximum allowed HDLC address (uint64_t in C++).
MAX_ADDRESS = 2**64 - 1


def escape(byte: int) -> int:
    """Escapes or unescapes a byte, which should have been preceeded by 0x7d."""
    return byte ^ 0x20


def frame_check_sequence(data: bytes) -> bytes:
    return zlib.crc32(data).to_bytes(4, 'little')


def encode_address(address: int) -> bytes:
    """Encodes an HDLC address as a one-terminated LSB varint."""
    result = bytearray()

    while True:
        result += bytes([(address & 0x7F) << 1])

        address >>= 7
        if address == 0:
            break

    result[-1] |= 0x1
    return result


def decode_address(frame: bytes) -> Tuple[int, int]:
    """Decodes an HDLC address from a frame, returning it and its size."""
    result = 0
    length = 0

    while length < len(frame):
        byte = frame[length]
        result |= (byte >> 1) << (length * 7)
        length += 1

        if byte & 0x1 == 0x1:
            break

    if result > MAX_ADDRESS:
        return -1, 0

    return result, length


class UFrameControl:
    def __init__(self, frame_type: int):
        self._data: bytes = bytes([0x03 | frame_type])

    @property
    def data(self):
        return self._data

    @classmethod
    def unnumbered_information(cls):
        return UFrameControl(0x00)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/update_bundle.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_software_update import tuf_pb2 as pw__software__update_dot_tuf__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_software_update/update_bundle.proto',
  package='pw.software_update',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&pw_software_update/update_bundle.proto\x12\x12pw.software_update\x1a\x1cpw_software_update/tuf.proto\"\xe5\x04\n\x0cUpdateBundle\x12L\n\x12timestamp_metadata\x18\x01 \x01(\x0b\x32+.pw.software_update.SignedTimestampMetadataH\x00\x88\x01\x01\x12J\n\x11snapshot_metadata\x18\x02 \x01(\x0b\x32*.pw.software_update.SignedSnapshotMetadataH\x01\x88\x01\x01\x12O\n\x10targets_metadata\x18\x03 \x03(\x0b\x32\x35.pw.software_update.UpdateBundle.TargetsMetadataEntry\x12M\n\x0ftarget_payloads\x18\x04 \x03(\x0b\x32\x34.pw.software_update.UpdateBundle.TargetPayloadsEntry\x12\x42\n\rroot_metadata\x18\x05 \x01(\x0b\x32&.pw.software_update.SignedRootMetadataH\x02\x88\x01\x01\x1a\x61\n\x14TargetsMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x38\n\x05value\x18\x02 \x01(\x0b\x32).pw.software_update.SignedTargetsMetadata:\x02\x38\x01\x1a\x35\n\x13TargetPayloadsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x42\x15\n\x13_timestamp_metadataB\x14\n\x12_snapshot_metadataB\x10\n\x0e_root_metadata\"\xe2\x01\n\x08Manifest\x12K\n\x10targets_metadata\x18\x01 \x03(\x0b\x32\x31.pw.software_update.Manifest.TargetsMetadataEntry\x12\x1a\n\ruser_manifest\x18\x02 \x01(\x0cH\x00\x88\x01\x01\x1a[\n\x14TargetsMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.pw.software_update.TargetsMetadata:\x02\x38\x01\x42\x10\n\x0e_user_manifestb\x06proto3'
  ,
  dependencies=[pw__software__update_dot_tuf__pb2.DESCRIPTOR,])




_UPDATEBUNDLE_TARGETSMETADATAENTRY = _descriptor.Descriptor(
  name='TargetsMetadataEntry',
  full_name='pw.software_update.UpdateBundle.TargetsMetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.software_update.UpdateBundle.TargetsMetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.software_update.UpdateBundle.TargetsMetadataEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=491,
  serialized_end=588,
)

_UPDATEBUNDLE_TARGETPAYLOADSENTRY = _descriptor.Descriptor(
  name='TargetPayloadsEntry',
  full_name='pw.software_update.UpdateBundle.TargetPayloadsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.software_update.UpdateBundle.TargetPayloadsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.software_update.UpdateBundle.TargetPayloadsEntry.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=590,
  serialized_end=643,
)

_UPDATEBUNDLE = _descriptor.Descriptor(
  name='UpdateBundle',
  full_name='pw.software_update.UpdateBundle',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp_metadata', full_name='pw.software_update.UpdateBundle.timestamp_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snapshot_metadata', full_name='pw.software_update.UpdateBundle.snapshot_metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='targets_metadata', full_name='pw.software_update.UpdateBundle.targets_metadata', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_payloads', full_name='pw.software_update.UpdateBundle.target_payloads', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='root_metadata', full_name='pw.software_update.UpdateBundle.root_metadata', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_UPDATEBUNDLE_TARGETSMETADATAENTRY, _UPDATEBUNDLE_TARGETPAYLOADSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_timestamp_metadata', full_name='pw.software_update.UpdateBundle._timestamp_metadata',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_snapshot_metadata', full_name='pw.software_update.UpdateBundle._snapshot_metadata',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_root_metadata', full_name='pw.software_update.UpdateBundle._root_metadata',
      index=2, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=93,
  serialized_end=706,
)


_MANIFEST_TARGETSMETADATAENTRY = _descriptor.Descriptor(
  name='TargetsMetadataEntry',
  full_name='pw.software_update.Manifest.TargetsMetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.software_update.Manifest.TargetsMetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.software_update.Manifest.TargetsMetadataEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=826,
  serialized_end=917,
)

_MANIFEST = _descriptor.Descriptor(
  name='Manifest',
  full_name='pw.software_update.Manifest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='targets_metadata', full_name='pw.software_update.Manifest.targets_metadata', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_manifest', full_name='pw.software_update.Manifest.user_manifest', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MANIFEST_TARGETSMETADATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_user_manifest', full_name='pw.software_update.Manifest._user_manifest',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=709,
  serialized_end=935,
)

_UPDATEBUNDLE_TARGETSMETADATAENTRY.fields_by_name['value'].message_type = pw__software__update_dot_tuf__pb2._SIGNEDTARGETSMETADATA
_UPDATEBUNDLE_TARGETSMETADATAENTRY.containing_type = _UPDATEBUNDLE
_UPDATEBUNDLE_TARGETPAYLOADSENTRY.containing_type = _UPDATEBUNDLE
_UPDATEBUNDLE.fields_by_name['timestamp_metadata'].message_type = pw__software__update_dot_tuf__pb2._SIGNEDTIMESTAMPMETADATA
_UPDATEBUNDLE.fields_by_name['snapshot_metadata'].message_type = pw__software__update_dot_tuf__pb2._SIGNEDSNAPSHOTMETADATA
_UPDATEBUNDLE.fields_by_name['targets_metadata'].message_type = _UPDATEBUNDLE_TARGETSMETADATAENTRY
_UPDATEBUNDLE.fields_by_name['target_payloads'].message_type = _UPDATEBUNDLE_TARGETPAYLOADSENTRY
_UPDATEBUNDLE.fields_by_name['root_metadata'].message_type = pw__software__update_dot_tuf__pb2._SIGNEDROOTMETADATA
_UPDATEBUNDLE.oneofs_by_name['_timestamp_metadata'].fields.append(
  _UPDATEBUNDLE.fields_by_name['timestamp_metadata'])
_UPDATEBUNDLE.fields_by_name['timestamp_metadata'].containing_oneof = _UPDATEBUNDLE.oneofs_by_name['_timestamp_metadata']
_UPDATEBUNDLE.oneofs_by_name['_snapshot_metadata'].fields.append(
  _UPDATEBUNDLE.fields_by_name['snapshot_metadata'])
_UPDATEBUNDLE.fields_by_name['snapshot_metadata'].containing_oneof = _UPDATEBUNDLE.oneofs_by_name['_snapshot_metadata']
_UPDATEBUNDLE.oneofs_by_name['_root_metadata'].fields.append(
  _UPDATEBUNDLE.fields_by_name['root_metadata'])
_UPDATEBUNDLE.fields_by_name['root_metadata'].containing_oneof = _UPDATEBUNDLE.oneofs_by_name['_root_metadata']
_MANIFEST_TARGETSMETADATAENTRY.fields_by_name['value'].message_type = pw__software__update_dot_tuf__pb2._TARGETSMETADATA
_MANIFEST_TARGETSMETADATAENTRY.containing_type = _MANIFEST
_MANIFEST.fields_by_name['targets_metadata'].message_type = _MANIFEST_TARGETSMETADATAENTRY
_MANIFEST.oneofs_by_name['_user_manifest'].fields.append(
  _MANIFEST.fields_by_name['user_manifest'])
_MANIFEST.fields_by_name['user_manifest'].containing_oneof = _MANIFEST.oneofs_by_name['_user_manifest']
DESCRIPTOR.message_types_by_name['UpdateBundle'] = _UPDATEBUNDLE
DESCRIPTOR.message_types_by_name['Manifest'] = _MANIFEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UpdateBundle = _reflection.GeneratedProtocolMessageType('UpdateBundle', (_message.Message,), {

  'TargetsMetadataEntry' : _reflection.GeneratedProtocolMessageType('TargetsMetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEBUNDLE_TARGETSMETADATAENTRY,
    '__module__' : 'pw_software_update.update_bundle_pb2'
    # @@protoc_insertion_point(class_scope:pw.software_update.UpdateBundle.TargetsMetadataEntry)
    })
  ,

  'TargetPayloadsEntry' : _reflection.GeneratedProtocolMessageType('TargetPayloadsEntry', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEBUNDLE_TARGETPAYLOADSENTRY,
    '__module__' : 'pw_software_update.update_bundle_pb2'
    # @@protoc_insertion_point(class_scope:pw.software_update.UpdateBundle.TargetPayloadsEntry)
    })
  ,
  'DESCRIPTOR' : _UPDATEBUNDLE,
  '__module__' : 'pw_software_update.update_bundle_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.UpdateBundle)
  })
_sym_db.RegisterMessage(UpdateBundle)
_sym_db.RegisterMessage(UpdateBundle.TargetsMetadataEntry)
_sym_db.RegisterMessage(UpdateBundle.TargetPayloadsEntry)

Manifest = _reflection.GeneratedProtocolMessageType('Manifest', (_message.Message,), {

  'TargetsMetadataEntry' : _reflection.GeneratedProtocolMessageType('TargetsMetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _MANIFEST_TARGETSMETADATAENTRY,
    '__module__' : 'pw_software_update.update_bundle_pb2'
    # @@protoc_insertion_point(class_scope:pw.software_update.Manifest.TargetsMetadataEntry)
    })
  ,
  'DESCRIPTOR' : _MANIFEST,
  '__module__' : 'pw_software_update.update_bundle_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.Manifest)
  })
_sym_db.RegisterMessage(Manifest)
_sym_db.RegisterMessage(Manifest.TargetsMetadataEntry)


_UPDATEBUNDLE_TARGETSMETADATAENTRY._options = None
_UPDATEBUNDLE_TARGETPAYLOADSENTRY._options = None
_MANIFEST_TARGETSMETADATAENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Container class for a single progress bar task."""

from dataclasses import dataclass
from typing import Optional

from prompt_toolkit.application import get_app_or_none
from prompt_toolkit.shortcuts.progress_bar import ProgressBarCounter


def _redraw_ui() -> None:
    """Signal the prompt_toolkit app to re-draw"""
    pt_app = get_app_or_none()
    if pt_app:
        pt_app.invalidate()


@dataclass
class ProgressBarTaskCounter:
    """Class to hold a single progress bar state."""

    name: str
    total: int
    count: int = 0
    completed: bool = False
    canceled: bool = False
    prompt_toolkit_counter: Optional[ProgressBarCounter] = None

    def mark_canceled(self):
        self.canceled = True
        self.prompt_toolkit_counter.stopped = True  # type: ignore

    def mark_completed(self):
        self.completed = True
        self.prompt_toolkit_counter.done = True  # type: ignore

    def check_completion(self) -> None:
        # Check for completion
        if self.count >= self.total:
            self.mark_completed()

    def stop_updating_prompt_toolkit_counter(self) -> None:
        """If count is over total, stop updating the prompt_toolkit ETA."""
        if self.count >= self.total:
            self.prompt_toolkit_counter.done = True  # type: ignore

    def update(self, count: int = 1) -> None:
        """Increment this counter."""
        self.count += count

        if self.prompt_toolkit_counter:
            self.prompt_toolkit_counter.items_completed += count
            self.stop_updating_prompt_toolkit_counter()
            _redraw_ui()

    def set_new_total(self, new_total: int) -> None:
        """Set a new total count."""
        self.count = new_total

        if self.prompt_toolkit_counter:
            self.prompt_toolkit_counter.items_completed = new_total
            self.stop_updating_prompt_toolkit_counter()
            _redraw_ui()



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Extracts a concise error from a bazel log."""

from pathlib import Path
import re
import sys


def parse_bazel_stdout(bazel_stdout: Path) -> str:
    """Extracts a concise error from a bazel log."""
    seen_error = False
    error_lines = []

    with bazel_stdout.open() as ins:
        for line in ins:
            # Trailing whitespace isn't significant, as it doesn't affect the
            # way the line shows up in the logs. However, leading whitespace may
            # be significant, especially for compiler error messages.
            line = line.rstrip()

            if re.match(r'^ERROR:', line):
                seen_error = True

            if seen_error:
                # Ignore long lines that just show the environment.
                if re.search(r' +[\w_]*(PATH|PWD)=.* \\', line):
                    continue

                # Ignore lines that only show bazel sandboxing.
                if line.strip().startswith(('(cd /', 'exec env')):
                    continue

                if line.strip().startswith('Use --sandbox_debug to see'):
                    continue

                if line.strip().startswith('# Configuration'):
                    continue

                # An "<ALLCAPS>:" line usually means compiler output is done
                # and useful compiler errors are complete.
                if re.match(r'^(?!ERROR)[A-Z]+:', line):
                    break

                error_lines.append(line)

    result = '\n'.join(error_lines)
    return re.sub(r'\n+', '\n', result)


if __name__ == '__main__':
    print(parse_bazel_stdout(Path(sys.argv[1])))



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generate and serialize update bundles."""

import argparse
import logging
import os
from pathlib import Path
import shutil
from typing import Dict, Iterable, Optional, Tuple

from pw_software_update import metadata
from pw_software_update.tuf_pb2 import SignedRootMetadata, SignedTargetsMetadata
from pw_software_update.update_bundle_pb2 import UpdateBundle

_LOG = logging.getLogger(__package__)


def targets_from_directory(
    root_dir: Path,
    exclude: Iterable[Path] = tuple(),
    remap_paths: Optional[Dict[Path, str]] = None,
) -> Dict[str, Path]:
    """Given a directory on dist, generate a dict of target names to files.

    Args:
      root_dir: Directory to crawl for targets.
      exclude: Paths relative to root_dir to exclude as targets.
      remap_paths: Custom target names to use for targets.

    Each file in the input directory will be read in as a target file, unless
    its path (relative to the TUF repo root) is among the excludes.

    Default behavior is to treat root_dir-relative paths as the strings to use
    as targets file names, but remapping can be used to change a target file
    name to any string. If some remappings are provided but a file is found that
    does not have a remapping, a warning will be logged. If a remap is declared
    for a file that does not exist, FileNotFoundError will be raised.
    """
    if not root_dir.is_dir():
        raise ValueError(
            f'Cannot generate TUF targets from {root_dir}; not a directory.'
        )
    targets = {}
    for path in root_dir.glob('**/*'):
        if path.is_dir():
            continue
        rel_path = path.relative_to(root_dir)
        if rel_path in exclude:
            continue
        target_name = str(rel_path.as_posix())
        if remap_paths:
            if rel_path in remap_paths:
                target_name = remap_paths[rel_path]
            else:
                _LOG.warning('Some remaps defined, but not "%s"', target_name)
        targets[target_name] = path

    if remap_paths is not None:
        for original_path, new_target_file_name in remap_paths.items():
            if new_target_file_name not in targets:
                raise FileNotFoundError(
                    f'Unable to remap "{original_path}" to'
                    f' "{new_target_file_name}"; file not found in root dir.'
                )

    return targets


def gen_empty_update_bundle(
    targets_metadata_version: int = metadata.DEFAULT_METADATA_VERSION,
) -> UpdateBundle:
    """Generates an empty bundle

    Given an optional target metadata version, generates an empty bundle.

    Args:
      targets_metadata_version: default set to 1

    Returns:
      UpdateBundle: empty bundle
    """

    targets_metadata = metadata.gen_targets_metadata(
        target_payloads={}, version=targets_metadata_version
    )
    unsigned_targets_metadata = SignedTargetsMetadata(
        serialized_targets_metadata=targets_metadata.SerializeToString()
    )

    return UpdateBundle(
        root_metadata=None,
        targets_metadata=dict(targets=unsigned_targets_metadata),
        target_payloads=None,
    )


def gen_unsigned_update_bundle(
    targets: Dict[Path, str],
    persist: Optional[Path] = None,
    targets_metadata_version: int = metadata.DEFAULT_METADATA_VERSION,
    root_metadata: Optional[SignedRootMetadata] = None,
) -> UpdateBundle:
    """Given a set of targets, generates an unsigned UpdateBundle.

    Args:
      targets: A dict mapping payload Paths to their target names.
      persist: If not None, persist the raw TUF repository to this directory.
      targets_metadata_version: version number for the targets metadata.
      root_metadata: Optional signed Root metadata.

    The input targets will be treated as an ephemeral TUF repository for the
    purposes of building an UpdateBundle instance. This approach differs
    slightly from the normal concept of a TUF repository, which is typically a
    directory on disk. For ease in debugging raw repository contents, the
    `persist` argument can be supplied. If a persist Path is supplied, the TUF
    repository will be persisted to disk at that location.

    NOTE: If path separator characters (like '/') are used in target names, then
    persisting the repository to disk via the 'persist' argument will create the
    corresponding directory structure.

    NOTE: If a root metadata is included, the client is expected to first
    upgrade its on-device trusted root metadata before verifying the rest of
    the bundle.
    """
    if persist:
        if persist.exists() and not persist.is_dir():
            raise ValueError(
                f'TUF repo cannot be persisted to "{persist}";'
                ' file exists and is not a directory.'
            )
        if persist.exists():
            shutil.rmtree(persist)

        os.makedirs(persist)

    target_payloads = {}
    for path, target_name in targets.items():
        target_payloads[target_name] = path.read_bytes()
        if persist:
            target_persist_path = persist / target_name
            os.makedirs(target_persist_path.parent, exist_ok=True)
            shutil.copy(path, target_persist_path)

    targets_metadata = metadata.gen_targets_metadata(
        target_payloads, version=targets_metadata_version
    )
    unsigned_targets_metadata = SignedTargetsMetadata(
        serialized_targets_metadata=targets_metadata.SerializeToString()
    )

    return UpdateBundle(
        root_metadata=root_metadata,
        targets_metadata=dict(targets=unsigned_targets_metadata),
        target_payloads=target_payloads,
    )


def parse_target_arg(target_arg: str) -> Tuple[Path, str]:
    """Parse an individual target string passed in to the --targets argument.

    Target strings take the following form:
      "FILE_PATH > TARGET_NAME"

    For example:
      "fw_images/main_image.bin > main"
    """
    try:
        file_path_str, target_name = target_arg.split('>')
        return Path(file_path_str.strip()), target_name.strip()
    except ValueError as err:
        raise ValueError(
            'Targets must be strings of the form:\n'
            '  "FILE_PATH > TARGET_NAME"'
        ) from err


def parse_args() -> argparse.Namespace:
    """Parse CLI arguments."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '-t',
        '--targets',
        type=str,
        nargs='+',
        required=True,
        help='Strings defining targets to bundle',
    )
    parser.add_argument(
        '-o',
        '--out',
        type=Path,
        required=True,
        help='Output path for serialized UpdateBundle',
    )
    parser.add_argument(
        '--persist',
        type=Path,
        default=None,
        help=(
            'If provided, TUF repo will be persisted to disk'
            ' at this path for debugging'
        ),
    )
    parser.add_argument(
        '--targets-metadata-version',
        type=int,
        default=metadata.DEFAULT_METADATA_VERSION,
        help='Version number for the targets metadata',
    )
    parser.add_argument(
        '--targets-metadata-version-file',
        type=Path,
        default=None,
        help='Read version number string from this file. When '
        'provided, content of this file supersede '
        '--targets-metadata-version',
    )
    parser.add_argument(
        '--signed-root-metadata',
        type=Path,
        default=None,
        help='Path to the signed Root metadata',
    )
    return parser.parse_args()


def main(
    targets: Iterable[str],
    out: Path,
    persist: Optional[Path] = None,
    targets_metadata_version: int = metadata.DEFAULT_METADATA_VERSION,
    targets_metadata_version_file: Optional[Path] = None,
    signed_root_metadata: Optional[Path] = None,
) -> None:
    """Generates an UpdateBundle and serializes it to disk."""
    target_dict = {}
    for target_arg in targets:
        path, target_name = parse_target_arg(target_arg)
        target_dict[path] = target_name

    root_metadata = None
    if signed_root_metadata:
        root_metadata = SignedRootMetadata.FromString(
            signed_root_metadata.read_bytes()
        )

    if targets_metadata_version_file:
        with targets_metadata_version_file.open() as version_file:
            targets_metadata_version = int(version_file.read().strip())

    bundle = gen_unsigned_update_bundle(
        target_dict, persist, targets_metadata_version, root_metadata
    )

    out.write_bytes(bundle.SerializeToString())


if __name__ == '__main__':
    logging.basicConfig()
    main(**vars(parse_args()))



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Python logging helper fuctions."""

import copy
from datetime import datetime
import json
import logging
import tempfile
from typing import Any, Dict, Iterable, Iterator, Optional


def all_loggers() -> Iterator[logging.Logger]:
    """Iterates over all loggers known to Python logging."""
    manager = logging.getLogger().manager  # type: ignore[attr-defined]

    for logger_name in manager.loggerDict:  # pylint: disable=no-member
        yield logging.getLogger(logger_name)


def create_temp_log_file(
    prefix: Optional[str] = None, add_time: bool = True
) -> str:
    """Create a unique tempfile for saving logs.

    Example format: /tmp/pw_console_2021-05-04_151807_8hem6iyq
    """
    if not prefix:
        prefix = str(__package__)

    # Grab the current system timestamp as a string.
    isotime = datetime.now().isoformat(sep="_", timespec="seconds")
    # Timestamp string should not have colons in it.
    isotime = isotime.replace(":", "")

    if add_time:
        prefix += f"_{isotime}"

    log_file_name = None
    with tempfile.NamedTemporaryFile(
        prefix=f"{prefix}_", delete=False
    ) as log_file:
        log_file_name = log_file.name

    return log_file_name


def set_logging_last_resort_file_handler(
    file_name: Optional[str] = None,
) -> None:
    log_file = file_name if file_name else create_temp_log_file()
    logging.lastResort = logging.FileHandler(log_file)


def disable_stdout_handlers(logger: logging.Logger) -> None:
    """Remove all stdout and stdout & stderr logger handlers."""
    for handler in copy.copy(logger.handlers):
        # Must use type() check here since this returns True:
        #   isinstance(logging.FileHandler, logging.StreamHandler)
        # pylint: disable=unidiomatic-typecheck
        if type(handler) == logging.StreamHandler:
            logger.removeHandler(handler)
        # pylint: enable=unidiomatic-typecheck


def setup_python_logging(
    last_resort_filename: Optional[str] = None,
    loggers_with_no_propagation: Optional[Iterable[logging.Logger]] = None,
) -> None:
    """Disable log handlers for full screen prompt_toolkit applications."""
    if not loggers_with_no_propagation:
        loggers_with_no_propagation = []
    disable_stdout_handlers(logging.getLogger())

    if logging.lastResort is not None:
        set_logging_last_resort_file_handler(last_resort_filename)

    for logger in list(all_loggers()):
        # Prevent stdout handlers from corrupting the prompt_toolkit UI.
        disable_stdout_handlers(logger)
        if logger in loggers_with_no_propagation:
            continue
        # Make sure all known loggers propagate to the root logger.
        logger.propagate = True

    # Prevent these loggers from propagating to the root logger.
    hidden_host_loggers = [
        "pw_console",
        "pw_console.plugins",
        # prompt_toolkit triggered debug log messages
        "prompt_toolkit",
        "prompt_toolkit.buffer",
        "parso.python.diff",
        "parso.cache",
        "pw_console.serial_debug_logger",
    ]
    for logger_name in hidden_host_loggers:
        logging.getLogger(logger_name).propagate = False

    # Set asyncio log level to WARNING
    logging.getLogger("asyncio").setLevel(logging.WARNING)

    # Always set DEBUG level for serial debug.
    logging.getLogger("pw_console.serial_debug_logger").setLevel(logging.DEBUG)


def log_record_to_json(record: logging.LogRecord) -> str:
    log_dict: Dict[str, Any] = {}
    log_dict["message"] = record.getMessage()
    log_dict["levelno"] = record.levelno
    log_dict["levelname"] = record.levelname
    log_dict["args"] = record.args

    if hasattr(record, "extra_metadata_fields") and (
        record.extra_metadata_fields  # type: ignore
    ):
        fields = record.extra_metadata_fields  # type: ignore
        log_dict["fields"] = {}
        for key, value in fields.items():
            if key == "msg":
                log_dict["message"] = value
                continue

            log_dict["fields"][key] = str(value)

    return json.dumps(log_dict)


class JsonLogFormatter(logging.Formatter):
    """Json Python logging Formatter

    Use this formatter to log pw_console messages to a file in json
    format. Column values normally shown in table view will be populated in the
    'fields' key.

    Example log entry:

    .. code-block:: json

       {
         "message": "System init",
         "levelno": 20,
         "levelname": "INF",
         "args": [
           "0:00",
           "pw_system ",
           "System init"
         ],
         "fields": {
           "module": "pw_system",
           "file": "pw_system/init.cc",
           "timestamp": "0:00"
         }
       }

    Example usage:

    .. code-block:: python

       import logging
       import pw_console.python_logging

       _DEVICE_LOG = logging.getLogger('rpc_device')

       json_filehandler = logging.FileHandler('logs.json', encoding='utf-8')
       json_filehandler.setLevel(logging.DEBUG)
       json_filehandler.setFormatter(
           pw_console.python_logging.JsonLogFormatter())
       _DEVICE_LOG.addHandler(json_filehandler)

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def format(self, record: logging.LogRecord) -> str:
        return log_record_to_json(record)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Pigweed Console progress bar task state."""

from contextvars import ContextVar
import copy
from dataclasses import dataclass, field
import signal
from typing import Dict, Optional, Union

from prompt_toolkit.application import get_app_or_none
from prompt_toolkit.shortcuts import ProgressBar
from prompt_toolkit.shortcuts.progress_bar import formatters

from pw_console.progress_bar.progress_bar_impl import (
    IterationsPerSecondIfNotHidden,
    ProgressBarImpl,
    TextIfNotHidden,
    TimeLeftIfNotHidden,
)
from pw_console.progress_bar.progress_bar_task_counter import (
    ProgressBarTaskCounter,
)
from pw_console.style import generate_styles

CUSTOM_FORMATTERS = [
    formatters.Label(suffix=': '),
    formatters.Rainbow(
        formatters.Bar(start='|Pigw', end='|', sym_a='e', sym_b='d!', sym_c=' ')
    ),
    formatters.Text(' '),
    formatters.Progress(),
    formatters.Text(' ['),
    formatters.Percentage(),
    formatters.Text('] in '),
    formatters.TimeElapsed(),
    TextIfNotHidden(' ('),
    IterationsPerSecondIfNotHidden(),
    TextIfNotHidden('/s, eta: '),
    TimeLeftIfNotHidden(),
    TextIfNotHidden(')'),
]


def prompt_toolkit_app_running() -> bool:
    existing_app = get_app_or_none()
    if existing_app:
        return True
    return False


@dataclass
class ProgressBarState:
    """Pigweed Console wide state for all repl progress bars.

    An instance of this class is intended to be a global variable."""

    tasks: Dict[str, ProgressBarTaskCounter] = field(default_factory=dict)
    instance: Optional[Union[ProgressBar, ProgressBarImpl]] = None

    def _install_sigint_handler(self) -> None:
        """Add ctrl-c handling if not running inside pw_console"""

        def handle_sigint(_signum, _frame):
            # Shut down the ProgressBar prompt_toolkit application
            prog_bar = self.instance
            if prog_bar is not None and hasattr(prog_bar, '__exit__'):
                prog_bar.__exit__()  # pylint: disable=unnecessary-dunder-call
            raise KeyboardInterrupt

        signal.signal(signal.SIGINT, handle_sigint)

    def startup_progress_bar_impl(self):
        prog_bar = self.instance
        if not prog_bar:
            if prompt_toolkit_app_running():
                prog_bar = ProgressBarImpl(
                    style=get_app_or_none().style, formatters=CUSTOM_FORMATTERS
                )
            else:
                self._install_sigint_handler()
                prog_bar = ProgressBar(
                    style=generate_styles(), formatters=CUSTOM_FORMATTERS
                )
                # Start the ProgressBar prompt_toolkit application in a separate
                # thread.
                prog_bar.__enter__()  # pylint: disable=unnecessary-dunder-call
            self.instance = prog_bar
        return self.instance

    def cleanup_finished_tasks(self) -> None:
        for task_name in copy.copy(list(self.tasks.keys())):
            task = self.tasks[task_name]
            if task.completed or task.canceled:
                ptc = task.prompt_toolkit_counter
                self.tasks.pop(task_name, None)
                if (
                    self.instance
                    and self.instance.counters
                    and ptc in self.instance.counters
                ):
                    self.instance.counters.remove(ptc)

    @property
    def all_tasks_complete(self) -> bool:
        tasks_complete = [
            task.completed or task.canceled
            for _task_name, task in self.tasks.items()
        ]
        self.cleanup_finished_tasks()
        return all(tasks_complete)

    def cancel_all_tasks(self):
        self.tasks = {}
        if self.instance is not None:
            self.instance.counters = []

    def get_container(self):
        prog_bar = self.instance
        if prog_bar is not None and hasattr(prog_bar, '__pt_container__'):
            return prog_bar.__pt_container__()
        return None


TASKS_CONTEXTVAR = ContextVar(
    'pw_console_progress_bar_tasks', default=ProgressBarState()
)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Bandwidth Monitor Toolbar"""

from prompt_toolkit.layout import WindowAlign

from pw_console.plugin_mixin import PluginMixin
from pw_console.widgets import ToolbarButton, WindowPaneToolbar
from pw_console.pyserial_wrapper import BANDWIDTH_HISTORY_CONTEXTVAR


class BandwidthToolbar(WindowPaneToolbar, PluginMixin):
    """Toolbar for displaying bandwidth history."""

    TOOLBAR_HEIGHT = 1

    def _update_toolbar_text(self):
        """Format toolbar text.

        This queries pyserial_wrapper's EventCountHistory context var to
        retrieve the byte count history for read, write and totals."""
        tokens = []
        self.plugin_logger.debug('BandwidthToolbar _update_toolbar_text')

        for count_name, events in self.history.items():
            tokens.extend(
                [
                    ('', '  '),
                    (
                        'class:theme-bg-active class:theme-fg-active',
                        ' {}: '.format(count_name.title()),
                    ),
                    (
                        'class:theme-bg-active class:theme-fg-cyan',
                        '{:.3f} '.format(events.last_count()),
                    ),
                    (
                        'class:theme-bg-active class:theme-fg-orange',
                        '{} '.format(events.display_unit_title),
                    ),
                ]
            )
            if count_name == 'total':
                tokens.append(
                    ('class:theme-fg-cyan', '{}'.format(events.sparkline()))
                )

        self.formatted_text = tokens

    def get_left_text_tokens(self):
        """Formatted text to display on the far left side."""
        return self.formatted_text

    def get_right_text_tokens(self):
        """Formatted text to display on the far right side."""
        return [('class:theme-fg-blue', 'Serial Bandwidth Usage ')]

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, center_section_align=WindowAlign.RIGHT, **kwargs
        )

        self.history = BANDWIDTH_HISTORY_CONTEXTVAR.get()
        self.show_toolbar = True
        self.formatted_text = []

        # Buttons for display in the center
        self.add_button(
            ToolbarButton(
                description='Refresh', mouse_handler=self._update_toolbar_text
            )
        )

        # Set plugin options
        self.background_task_update_count: int = 0
        self.plugin_init(
            plugin_callback=self._background_task,
            plugin_callback_frequency=1.0,
            plugin_logger_name='pw_console_bandwidth_toolbar',
        )

    def _background_task(self) -> bool:
        self.background_task_update_count += 1
        self._update_toolbar_text()
        self.plugin_logger.debug(
            'BandwidthToolbar Scheduled Update: #%s',
            self.background_task_update_count,
        )
        return True



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""LogStore saves logs and acts as a Python logging handler."""

from __future__ import annotations
import collections
import logging
from datetime import datetime
from typing import Dict, List, Optional, TYPE_CHECKING

from pw_cli.color import colors as pw_cli_colors

from pw_console.console_prefs import ConsolePrefs
from pw_console.log_line import LogLine
from pw_console.text_formatting import strip_ansi
from pw_console.widgets.table import TableView

if TYPE_CHECKING:
    from pw_console.log_view import LogView


class LogStore(logging.Handler):
    """Pigweed Console logging handler.

    This is a `Python logging.Handler
    <https://docs.python.org/3/library/logging.html#handler-objects>`_ class
    used to store logs for display in the pw_console user interface.

    You may optionally add this as a handler to an existing logger
    instances. This will be required if logs need to be captured for display in
    the pw_console UI before the user interface is running.

    Example usage:

    .. code-block:: python

        import logging

        from pw_console import PwConsoleEmbed, LogStore

        _DEVICE_LOG = logging.getLogger('usb_gadget')

        # Create a log store and add as a handler.
        device_log_store = LogStore()
        _DEVICE_LOG.addHander(device_log_store)

        # Start communication with your device here, before embedding
        # pw_console.

        # Create the pw_console embed instance
        console = PwConsoleEmbed(
            global_vars=globals(),
            local_vars=locals(),
            loggers={
                'Host Logs': [
                    logging.getLogger(__package__),
                    logging.getLogger(__name__),
                ],
                # Set the LogStore as the value of this logger window.
                'Device Logs': device_log_store,
            },
            app_title='My Awesome Console',
        )

        console.setup_python_logging()
        console.embed()
    """

    def __init__(self, prefs: Optional[ConsolePrefs] = None):
        """Initializes the LogStore instance."""

        # ConsolePrefs may not be passed on init. For example, if the user is
        # creating a LogStore to capture log messages before console startup.
        if not prefs:
            prefs = ConsolePrefs(
                project_file=False, project_user_file=False, user_file=False
            )
        self.prefs = prefs
        # Log storage deque for fast addition and deletion from the beginning
        # and end of the iterable.
        self.logs: collections.deque = collections.deque()

        # Only allow this many log lines in memory.
        self.max_history_size: int = 1000000

        # Counts of logs per python logger name
        self.channel_counts: Dict[str, int] = {}
        # Widths of each logger prefix string. For example: the character length
        # of the timestamp string.
        self.channel_formatted_prefix_widths: Dict[str, int] = {}
        # Longest of the above prefix widths.
        self.longest_channel_prefix_width = 0

        self.table: TableView = TableView(prefs=self.prefs)

        # Erase existing logs.
        self.clear_logs()

        # List of viewers that should be notified on new log line arrival.
        self.registered_viewers: List['LogView'] = []

        super().__init__()

        # Set formatting after logging.Handler init.
        self.set_formatting()

    def set_prefs(self, prefs: ConsolePrefs) -> None:
        """Set the ConsolePrefs for this LogStore."""
        self.prefs = prefs
        self.table.set_prefs(prefs)

    def register_viewer(self, viewer: 'LogView') -> None:
        """Register this LogStore with a LogView."""
        self.registered_viewers.append(viewer)

    def set_formatting(self) -> None:
        """Setup log formatting."""
        # Copy of pw_cli log formatter
        colors = pw_cli_colors(True)
        timestamp_prefix = colors.black_on_white('%(asctime)s') + ' '
        timestamp_format = '%Y%m%d %H:%M:%S'
        format_string = timestamp_prefix + '%(levelname)s %(message)s'
        formatter = logging.Formatter(format_string, timestamp_format)

        self.setLevel(logging.DEBUG)
        self.setFormatter(formatter)

        # Update log time character width.
        example_time_string = datetime.now().strftime(timestamp_format)
        self.table.column_widths['time'] = len(example_time_string)

    def clear_logs(self):
        """Erase all stored pane lines."""
        self.logs = collections.deque()
        self.channel_counts = {}
        self.channel_formatted_prefix_widths = {}
        self.line_index = 0

    def get_channel_counts(self):
        """Return the seen channel log counts for this conatiner."""
        return ', '.join(
            [f'{name}: {count}' for name, count in self.channel_counts.items()]
        )

    def get_total_count(self):
        """Total size of the logs store."""
        return len(self.logs)

    def get_last_log_index(self):
        """Last valid index of the logs."""
        # Subtract 1 since self.logs is zero indexed.
        total = self.get_total_count()
        return 0 if total < 0 else total - 1

    def _update_log_prefix_width(self, record: logging.LogRecord):
        """Save the formatted prefix width if this is a new logger channel
        name."""
        if self.formatter and (
            record.name not in self.channel_formatted_prefix_widths.keys()
        ):
            # Find the width of the formatted timestamp and level
            format_string = (
                self.formatter._fmt  # pylint: disable=protected-access
            )

            # There may not be a _fmt defined.
            if not format_string:
                return

            format_without_message = format_string.replace('%(message)s', '')
            # If any other style parameters are left, get the width of them.
            if (
                format_without_message
                and 'asctime' in format_without_message
                and 'levelname' in format_without_message
            ):
                formatted_time_and_level = format_without_message % dict(
                    asctime=record.asctime, levelname=record.levelname
                )

                # Delete ANSI escape sequences.
                ansi_stripped_time_and_level = strip_ansi(
                    formatted_time_and_level
                )

                self.channel_formatted_prefix_widths[record.name] = len(
                    ansi_stripped_time_and_level
                )
            else:
                self.channel_formatted_prefix_widths[record.name] = 0

            # Set the max width of all known formats so far.
            self.longest_channel_prefix_width = max(
                self.channel_formatted_prefix_widths.values()
            )

    def _append_log(self, record: logging.LogRecord):
        """Add a new log event."""
        # Format incoming log line.
        formatted_log = self.format(record)
        ansi_stripped_log = strip_ansi(formatted_log)
        # Save this log.
        self.logs.append(
            LogLine(
                record=record,
                formatted_log=formatted_log,
                ansi_stripped_log=ansi_stripped_log,
            )
        )
        # Increment this logger count
        self.channel_counts[record.name] = (
            self.channel_counts.get(record.name, 0) + 1
        )

        # TODO(b/235271486): Revisit calculating prefix widths automatically
        # when line wrapping indentation is supported.
        # Set the prefix width to 0
        self.channel_formatted_prefix_widths[record.name] = 0

        # Parse metadata fields
        self.logs[-1].update_metadata()

        # Check for bigger column widths.
        self.table.update_metadata_column_widths(self.logs[-1])

    def emit(self, record) -> None:
        """Process a new log record.

        This defines the logging.Handler emit() fuction which is called by
        logging.Handler.handle() We don't implement handle() as it is done in
        the parent class with thread safety and filters applied.
        """
        self._append_log(record)
        # Notify viewers of new logs
        for viewer in self.registered_viewers:
            viewer.new_logs_arrived()

    def render_table_header(self):
        """Get pre-formatted table header."""
        return self.table.formatted_header()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2021 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class LogEntry(google.protobuf.message.Message):
    """A log message and metadata. Logs come in a few different forms:

     1. A tokenized log message (recommended for production)
     2. A non-tokenized log message (good for development)
     3. A "log missed" tombstone, indicating that some logs were dropped

    Size analysis for tokenized log messages, including each field's proto tag:

     - message     - 6-12 bytes; depending on number and value of arguments
     - line_level  - 3 bytes; 4 bytes if line > 2048 (uncommon)
     - timestamp   - 3 bytes; assuming delta encoding
     - thread      - 2-6 bytes; depending on whether value is a token or string

    Adding the fields gives the total proto message size:

       6-12 bytes - log
       9-15 bytes - log + level + line
      12-18 bytes - log + level + line + timestamp

    An analysis of a project's log token database revealed the following
    distribution of the number of arguments to log messages:

      # args   # messages
        0         2,700
        1         2,400
        2         1,200
        3+        1,000

    Note: The below proto makes some compromises compared to what one might
    expect for a "clean" proto design, in order to shave bytes off of the
    messages. It is critical that the log messages are as small as possible to
    enable storing more logs in limited memory. This is why, for example, there
    is no separate "DroppedLog" type, or a "TokenizedLog" and "StringLog", which
    would add at least 2 extra bytes per message
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGE_FIELD_NUMBER: builtins.int
    LINE_LEVEL_FIELD_NUMBER: builtins.int
    FLAGS_FIELD_NUMBER: builtins.int
    TIMESTAMP_FIELD_NUMBER: builtins.int
    TIME_SINCE_LAST_ENTRY_FIELD_NUMBER: builtins.int
    DROPPED_FIELD_NUMBER: builtins.int
    MODULE_FIELD_NUMBER: builtins.int
    FILE_FIELD_NUMBER: builtins.int
    THREAD_FIELD_NUMBER: builtins.int
    message: builtins.bytes
    """The log message, which may be tokenized.

    If tokenized logging is used, implementations may encode metadata in the
    log message rather than as separate proto fields. This reduces the size of
    the protobuf with no overhead.

    The standard format for encoding metadata in the log message is defined by
    the pw_log_tokenized module. The message and metadata are encoded as
    key-value pairs using ■ and ♦ as delimiters. For example:

     ■msg♦This is the log message: %d■module♦wifi■file♦../path/to/file.cc

    See http://pigweed.dev/pw_log_tokenized for full details. When
    pw_log_tokenized is used, this metadata is automatically included as
    described.

    The level and flags are not included since they may be runtime values and
    thus cannot always be tokenized. The line number is not included because
    line numbers change frequently and a new token is created for each line.

    Size analysis when tokenized:

      tag+wire = 1 byte
      size     = 1 byte; payload will almost always be < 127 bytes
      payload  = N bytes; typically 4-10 in practice

    Total: 2 + N ~= 6-12 bytes
    """
    line_level: builtins.int
    """Packed log level and line number. Structure:

      Level: Bottom 3 bits; level = line_level & 0x7
      Line: Remaining bits; line = (line_level >> 3)

    Note: This packing saves two bytes per log message in most cases compared
    to having line and level separately; and is zero-cost if the log backend
    omits the line number.
    """
    flags: builtins.int
    """Some log messages have flags to indicate attributes such as whether they
    are from an assert or if they contain PII. The particular flags are
    product- and implementation-dependent.
    """
    timestamp: builtins.int
    """The absolute timestamp in implementation-defined ticks. Applications
    determine how to interpret this on the receiving end. In the simplest
    case, these ticks might be milliseconds or microseconds since boot.
    Applications could also access clock information out-of-band with a
    ClockParameters protobuf.
    """
    time_since_last_entry: builtins.int
    """Time since the last entry in implementation-defined ticks, as for the
    timestamp field. This enables delta encoding when batching entries
    together.

    Size analysis for this field including tag and varint, assuming 1 kHz
    ticks:

              < 127 ms gap == 127 ms      ==  7 bits == 2 bytes
           < 16,000 ms gap ==  16 seconds == 14 bits == 3 bytes
        < 2,000,000 ms gap ==  35 minutes == 21 bits == 4 bytes
      < 300,000,000 ms gap ==  74 hours   == 28 bits == 5 bytes

    Log bursts will thus consume just 2 bytes (tag + up to 127ms delta) for
    the timestamp, which is a good improvement over an absolute timestamp.
    """
    dropped: builtins.int
    """When the log buffers are full but more logs come in, the logs are counted
    and a special log message is omitted with only counts for the number of
    messages dropped.
    """
    module: builtins.bytes
    """The PW_LOG_MODULE_NAME for this log message."""
    file: builtins.bytes
    """The file path where this log was created, if not encoded in the message."""
    thread: builtins.bytes
    """The task or thread name that created the log message. If the log was not
    created on a thread, it should use a name appropriate to that context.
    """
    def __init__(
        self,
        *,
        message: builtins.bytes | None = ...,
        line_level: builtins.int = ...,
        flags: builtins.int = ...,
        timestamp: builtins.int = ...,
        time_since_last_entry: builtins.int = ...,
        dropped: builtins.int = ...,
        module: builtins.bytes = ...,
        file: builtins.bytes = ...,
        thread: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_message", b"_message", "message", b"message", "time", b"time", "time_since_last_entry", b"time_since_last_entry", "timestamp", b"timestamp"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_message", b"_message", "dropped", b"dropped", "file", b"file", "flags", b"flags", "line_level", b"line_level", "message", b"message", "module", b"module", "thread", b"thread", "time", b"time", "time_since_last_entry", b"time_since_last_entry", "timestamp", b"timestamp"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_message", b"_message"]) -> typing_extensions.Literal["message"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["time", b"time"]) -> typing_extensions.Literal["timestamp", "time_since_last_entry"] | None: ...

global___LogEntry = LogEntry

class LogRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___LogRequest = LogRequest

class LogEntries(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENTRIES_FIELD_NUMBER: builtins.int
    FIRST_ENTRY_SEQUENCE_ID_FIELD_NUMBER: builtins.int
    @property
    def entries(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___LogEntry]: ...
    first_entry_sequence_id: builtins.int
    def __init__(
        self,
        *,
        entries: collections.abc.Iterable[global___LogEntry] | None = ...,
        first_entry_sequence_id: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["entries", b"entries", "first_entry_sequence_id", b"first_entry_sequence_id"]) -> None: ...

global___LogEntries = LogEntries

class FilterRule(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Level:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _LevelEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[FilterRule._Level.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        ANY_LEVEL: FilterRule._Level.ValueType  # 0
        DEBUG_LEVEL: FilterRule._Level.ValueType  # 1
        INFO_LEVEL: FilterRule._Level.ValueType  # 2
        WARN_LEVEL: FilterRule._Level.ValueType  # 3
        ERROR_LEVEL: FilterRule._Level.ValueType  # 4
        CRITICAL_LEVEL: FilterRule._Level.ValueType  # 5
        FATAL_LEVEL: FilterRule._Level.ValueType  # 7

    class Level(_Level, metaclass=_LevelEnumTypeWrapper):
        """Log level values match pw_log/levels.h. Enum names avoid collissions with
        possible macros.
        """

    ANY_LEVEL: FilterRule.Level.ValueType  # 0
    DEBUG_LEVEL: FilterRule.Level.ValueType  # 1
    INFO_LEVEL: FilterRule.Level.ValueType  # 2
    WARN_LEVEL: FilterRule.Level.ValueType  # 3
    ERROR_LEVEL: FilterRule.Level.ValueType  # 4
    CRITICAL_LEVEL: FilterRule.Level.ValueType  # 5
    FATAL_LEVEL: FilterRule.Level.ValueType  # 7

    class _Action:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ActionEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[FilterRule._Action.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        INACTIVE: FilterRule._Action.ValueType  # 0
        """Ignore the rule entirely."""
        KEEP: FilterRule._Action.ValueType  # 1
        """Keep the log entry if all conditions are met."""
        DROP: FilterRule._Action.ValueType  # 2
        """Drop the log entry if all conditions are met"""

    class Action(_Action, metaclass=_ActionEnumTypeWrapper):
        """Action to take if all conditions are met and rule is not inactive."""

    INACTIVE: FilterRule.Action.ValueType  # 0
    """Ignore the rule entirely."""
    KEEP: FilterRule.Action.ValueType  # 1
    """Keep the log entry if all conditions are met."""
    DROP: FilterRule.Action.ValueType  # 2
    """Drop the log entry if all conditions are met"""

    LEVEL_GREATER_THAN_OR_EQUAL_FIELD_NUMBER: builtins.int
    MODULE_EQUALS_FIELD_NUMBER: builtins.int
    ANY_FLAGS_SET_FIELD_NUMBER: builtins.int
    ACTION_FIELD_NUMBER: builtins.int
    THREAD_EQUALS_FIELD_NUMBER: builtins.int
    level_greater_than_or_equal: global___FilterRule.Level.ValueType
    """Condition 1: log.level >= level_greater_than_or_equal."""
    module_equals: builtins.bytes
    """Condition 2: (module_equals.size() == 0) || (log.module == module_equals);"""
    any_flags_set: builtins.int
    """Condition 3: (any_flags_set == 0) || (log.flags & any_flags_set) != 0))"""
    action: global___FilterRule.Action.ValueType
    thread_equals: builtins.bytes
    """Condition 4: (thread_equals.size() == 0 || log.thread == thread_equals)."""
    def __init__(
        self,
        *,
        level_greater_than_or_equal: global___FilterRule.Level.ValueType = ...,
        module_equals: builtins.bytes = ...,
        any_flags_set: builtins.int = ...,
        action: global___FilterRule.Action.ValueType = ...,
        thread_equals: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["action", b"action", "any_flags_set", b"any_flags_set", "level_greater_than_or_equal", b"level_greater_than_or_equal", "module_equals", b"module_equals", "thread_equals", b"thread_equals"]) -> None: ...

global___FilterRule = FilterRule

class Filter(google.protobuf.message.Message):
    """A filter is a series of rules. First matching rule wins."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RULE_FIELD_NUMBER: builtins.int
    @property
    def rule(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FilterRule]: ...
    def __init__(
        self,
        *,
        rule: collections.abc.Iterable[global___FilterRule] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["rule", b"rule"]) -> None: ...

global___Filter = Filter

class SetFilterRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILTER_ID_FIELD_NUMBER: builtins.int
    FILTER_FIELD_NUMBER: builtins.int
    filter_id: builtins.bytes
    """A filter can be identified by a human readable string, token, or number."""
    @property
    def filter(self) -> global___Filter: ...
    def __init__(
        self,
        *,
        filter_id: builtins.bytes = ...,
        filter: global___Filter | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["filter", b"filter"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["filter", b"filter", "filter_id", b"filter_id"]) -> None: ...

global___SetFilterRequest = SetFilterRequest

class GetFilterRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILTER_ID_FIELD_NUMBER: builtins.int
    filter_id: builtins.bytes
    def __init__(
        self,
        *,
        filter_id: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["filter_id", b"filter_id"]) -> None: ...

global___GetFilterRequest = GetFilterRequest

class FilterIdListRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___FilterIdListRequest = FilterIdListRequest

class FilterIdListResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILTER_ID_FIELD_NUMBER: builtins.int
    @property
    def filter_id(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(
        self,
        *,
        filter_id: collections.abc.Iterable[builtins.bytes] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["filter_id", b"filter_id"]) -> None: ...

global___FilterIdListResponse = FilterIdListResponse



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Evaluates target expressions within a GN build context."""

import argparse
from dataclasses import dataclass
import enum
import logging
import os
import re
import sys
from pathlib import Path
from typing import (
    Callable,
    Dict,
    List,
    Iterable,
    Iterator,
    NamedTuple,
    Optional,
    Tuple,
)

_LOG = logging.getLogger(__name__)


def abspath(path: Path) -> Path:
    """Turns a path into an absolute path, not resolving symlinks."""
    return Path(os.path.abspath(path))


class GnPaths(NamedTuple):
    """The set of paths needed to resolve GN paths to filesystem paths."""

    root: Path
    build: Path
    cwd: Path

    # Toolchain label or '' if using the default toolchain
    toolchain: str

    def resolve(self, gn_path: str) -> Path:
        """Resolves a GN path to a filesystem path."""
        if gn_path.startswith('//'):
            return abspath(self.root.joinpath(gn_path.lstrip('/')))

        return abspath(self.cwd.joinpath(gn_path))

    def resolve_paths(self, gn_paths: str, sep: str = ';') -> str:
        """Resolves GN paths to filesystem paths in a delimited string."""
        return sep.join(str(self.resolve(path)) for path in gn_paths.split(sep))


@dataclass(frozen=True)
class Label:
    """Represents a GN label."""

    name: str
    dir: Path
    relative_dir: Path
    toolchain: Optional['Label']
    out_dir: Path
    gen_dir: Path

    def __init__(self, paths: GnPaths, label: str):
        # Use this lambda to set attributes on this frozen dataclass.
        set_attr = lambda attr, val: object.__setattr__(self, attr, val)

        # Handle explicitly-specified toolchains
        if label.endswith(')'):
            label, toolchain = label[:-1].rsplit('(', 1)
        else:
            # Prevent infinite recursion for toolchains
            toolchain = paths.toolchain if paths.toolchain != label else ''

        set_attr('toolchain', Label(paths, toolchain) if toolchain else None)

        # Split off the :target, if provided, or use the last part of the path.
        try:
            directory, name = label.rsplit(':', 1)
        except ValueError:
            directory, name = label, label.rsplit('/', 1)[-1]

        set_attr('name', name)

        # Resolve the directory to an absolute path
        set_attr('dir', paths.resolve(directory))
        set_attr('relative_dir', self.dir.relative_to(abspath(paths.root)))

        set_attr(
            'out_dir',
            paths.build / self.toolchain_name() / 'obj' / self.relative_dir,
        )
        set_attr(
            'gen_dir',
            paths.build / self.toolchain_name() / 'gen' / self.relative_dir,
        )

    def gn_label(self) -> str:
        label = f'//{self.relative_dir.as_posix()}:{self.name}'
        return f'{label}({self.toolchain!r})' if self.toolchain else label

    def toolchain_name(self) -> str:
        return self.toolchain.name if self.toolchain else ''

    def __repr__(self) -> str:
        return self.gn_label()


class _Artifact(NamedTuple):
    path: Path
    variables: Dict[str, str]


# Matches a non-phony build statement.
_GN_NINJA_BUILD_STATEMENT = re.compile(r'^build (.+):[ \n](?!phony\b)')

_OBJECTS_EXTENSIONS = ('.o',)

# Extensions used for compilation artifacts.
_MAIN_ARTIFACTS = '', '.elf', '.a', '.so', '.dylib', '.exe', '.lib', '.dll'


def _get_artifact(entries: List[str]) -> _Artifact:
    """Attempts to resolve which artifact to use if there are multiple.

    Selects artifacts based on extension. This will not work if a toolchain
    creates multiple compilation artifacts from one command (e.g. .a and .elf).
    """
    assert entries, "There should be at least one entry here!"

    if len(entries) == 1:
        return _Artifact(Path(entries[0]), {})

    filtered = [p for p in entries if Path(p).suffix in _MAIN_ARTIFACTS]

    if len(filtered) == 1:
        return _Artifact(Path(filtered[0]), {})

    raise ExpressionError(
        f'Expected 1, but found {len(filtered)} artifacts, after filtering for '
        f'extensions {", ".join(repr(e) for e in _MAIN_ARTIFACTS)}: {entries}'
    )


def _parse_build_artifacts(fd) -> Iterator[_Artifact]:
    """Partially parses the build statements in a Ninja file."""
    lines = iter(fd)

    def next_line():
        try:
            return next(lines)
        except StopIteration:
            return None

    # Serves as the parse state (only two states)
    artifact: Optional[_Artifact] = None

    line = next_line()

    while line is not None:
        if artifact:
            if line.startswith('  '):  # build variable statements are indented
                key, value = (a.strip() for a in line.split('=', 1))
                artifact.variables[key] = value
                line = next_line()
            else:
                yield artifact
                artifact = None
        else:
            match = _GN_NINJA_BUILD_STATEMENT.match(line)
            if match:
                artifact = _get_artifact(match.group(1).split())

            line = next_line()

    if artifact:
        yield artifact


def _search_target_ninja(
    ninja_file: Path, target: Label
) -> Tuple[Optional[Path], List[Path]]:
    """Parses the main output file and object files from <target>.ninja."""

    artifact: Optional[Path] = None
    objects: List[Path] = []

    _LOG.debug('Parsing target Ninja file %s for %s', ninja_file, target)

    with ninja_file.open() as fd:
        for path, _ in _parse_build_artifacts(fd):
            # Older GN used .stamp files when there is no build artifact.
            if path.suffix == '.stamp':
                continue

            if str(path).endswith(_OBJECTS_EXTENSIONS):
                objects.append(Path(path))
            else:
                assert not artifact, f'Multiple artifacts for {target}!'
                artifact = Path(path)

    return artifact, objects


def _search_toolchain_ninja(
    ninja_file: Path, paths: GnPaths, target: Label
) -> Optional[Path]:
    """Searches the toolchain.ninja file for outputs from the provided target.

    Files created by an action appear in toolchain.ninja instead of in their own
    <target>.ninja. If the specified target has a single output file in
    toolchain.ninja, this function returns its path.
    """

    _LOG.debug('Searching toolchain Ninja file %s for %s', ninja_file, target)

    # Older versions of GN used a .stamp file to signal completion of a target.
    stamp_dir = target.out_dir.relative_to(paths.build).as_posix()
    stamp_tool = 'stamp'
    if target.toolchain_name() != '':
        stamp_tool = f'{target.toolchain_name()}_stamp'
    stamp_statement = f'build {stamp_dir}/{target.name}.stamp: {stamp_tool} '

    # Newer GN uses a phony Ninja target to signal completion of a target.
    phony_dir = Path(
        target.toolchain_name(), 'phony', target.relative_dir
    ).as_posix()
    phony_statement = f'build {phony_dir}/{target.name}: phony '

    with ninja_file.open() as fd:
        for line in fd:
            for statement in (phony_statement, stamp_statement):
                if line.startswith(statement):
                    output_files = line[len(statement) :].strip().split()
                    if len(output_files) == 1:
                        return Path(output_files[0])

                    break

    return None


def _search_ninja_files(
    paths: GnPaths, target: Label
) -> Tuple[bool, Optional[Path], List[Path]]:
    ninja_file = target.out_dir / f'{target.name}.ninja'
    if ninja_file.exists():
        return (True, *_search_target_ninja(ninja_file, target))

    ninja_file = paths.build / target.toolchain_name() / 'toolchain.ninja'
    if ninja_file.exists():
        return True, _search_toolchain_ninja(ninja_file, paths, target), []

    return False, None, []


@dataclass(frozen=True)
class TargetInfo:
    """Provides information about a target parsed from a .ninja file."""

    label: Label
    generated: bool  # True if the Ninja files for this target were generated.
    artifact: Optional[Path]
    object_files: Tuple[Path]

    def __init__(self, paths: GnPaths, target: str):
        object.__setattr__(self, 'label', Label(paths, target))

        generated, artifact, objects = _search_ninja_files(paths, self.label)

        object.__setattr__(self, 'generated', generated)
        object.__setattr__(self, 'artifact', artifact)
        object.__setattr__(self, 'object_files', tuple(objects))

    def __repr__(self) -> str:
        return repr(self.label)


class ExpressionError(Exception):
    """An error occurred while parsing an expression."""


class _ArgAction(enum.Enum):
    APPEND = 0
    OMIT = 1
    EMIT_NEW = 2


class _Expression:
    def __init__(self, match: re.Match, ending: int):
        self._match = match
        self._ending = ending

    @property
    def string(self):
        return self._match.string

    @property
    def end(self) -> int:
        return self._ending + len(_ENDING)

    def contents(self) -> str:
        return self.string[self._match.end() : self._ending]

    def expression(self) -> str:
        return self.string[self._match.start() : self.end]


_Actions = Iterator[Tuple[_ArgAction, str]]


def _target_file(paths: GnPaths, expr: _Expression) -> _Actions:
    target = TargetInfo(paths, expr.contents())

    if not target.generated:
        raise ExpressionError(f'Target {target} has not been generated by GN!')

    if target.artifact is None:
        raise ExpressionError(f'Target {target} has no output file!')

    yield _ArgAction.APPEND, str(target.artifact)


def _target_file_if_exists(paths: GnPaths, expr: _Expression) -> _Actions:
    target = TargetInfo(paths, expr.contents())

    if target.generated:
        if target.artifact is None:
            raise ExpressionError(f'Target {target} has no output file!')

        if paths.build.joinpath(target.artifact).exists():
            yield _ArgAction.APPEND, str(target.artifact)
            return

    yield _ArgAction.OMIT, ''


def _target_objects(paths: GnPaths, expr: _Expression) -> _Actions:
    if expr.expression() != expr.string:
        raise ExpressionError(
            f'The expression "{expr.expression()}" in "{expr.string}" may '
            'expand to multiple arguments, so it cannot be used alongside '
            'other text or expressions'
        )

    target = TargetInfo(paths, expr.contents())
    if not target.generated:
        raise ExpressionError(f'Target {target} has not been generated by GN!')

    for obj in target.object_files:
        yield _ArgAction.EMIT_NEW, str(obj)


# TODO(b/234886742): Replace expressions with native GN features when possible.
_FUNCTIONS: Dict['str', Callable[[GnPaths, _Expression], _Actions]] = {
    'TARGET_FILE': _target_file,
    'TARGET_FILE_IF_EXISTS': _target_file_if_exists,
    'TARGET_OBJECTS': _target_objects,
}

_START_EXPRESSION = re.compile(fr'<({"|".join(_FUNCTIONS)})\(')
_ENDING = ')>'


def _expand_arguments(paths: GnPaths, string: str) -> _Actions:
    pos = 0

    for match in _START_EXPRESSION.finditer(string):
        if pos != match.start():
            yield _ArgAction.APPEND, string[pos : match.start()]

        ending = string.find(_ENDING, match.end())
        if ending == -1:
            raise ExpressionError(
                f'Parse error: no terminating "{_ENDING}" '
                f'was found for "{string[match.start():]}"'
            )

        expression = _Expression(match, ending)
        yield from _FUNCTIONS[match.group(1)](paths, expression)

        pos = expression.end

    if pos < len(string):
        yield _ArgAction.APPEND, string[pos:]


def expand_expressions(paths: GnPaths, arg: str) -> Iterable[str]:
    """Expands <FUNCTION(...)> expressions; yields zero or more arguments."""
    if arg == '':
        return ['']

    expanded_args: List[List[str]] = [[]]

    for action, piece in _expand_arguments(paths, arg):
        if action is _ArgAction.OMIT:
            return []

        expanded_args[-1].append(piece)
        if action is _ArgAction.EMIT_NEW:
            expanded_args.append([])

    return (''.join(arg) for arg in expanded_args if arg)


def _parse_args() -> argparse.Namespace:
    file_pair = lambda s: tuple(Path(p) for p in s.split(':'))

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--gn-root',
        type=Path,
        required=True,
        help=(
            'Path to the root of the GN tree; '
            'value of rebase_path("//", root_build_dir)'
        ),
    )
    parser.add_argument(
        '--current-path',
        type=Path,
        required=True,
        help='Value of rebase_path(".", root_build_dir)',
    )
    parser.add_argument(
        '--default-toolchain', required=True, help='Value of default_toolchain'
    )
    parser.add_argument(
        '--current-toolchain', required=True, help='Value of current_toolchain'
    )
    parser.add_argument(
        'files',
        metavar='FILE',
        nargs='+',
        type=file_pair,
        help='Pairs of src:dest files to scan for expressions to evaluate',
    )
    return parser.parse_args()


def _resolve_expressions_in_file(src: Path, dst: Path, paths: GnPaths):
    dst.write_text(''.join(expand_expressions(paths, src.read_text())))


def main(
    gn_root: Path,
    current_path: Path,
    default_toolchain: str,
    current_toolchain: str,
    files: Iterable[Tuple[Path, Path]],
) -> int:
    """Evaluates GN target expressions within a list of files.

    Modifies the files in-place with their resolved contents.
    """
    tool = current_toolchain if current_toolchain != default_toolchain else ''
    paths = GnPaths(
        root=abspath(gn_root),
        build=Path.cwd(),
        cwd=abspath(current_path),
        toolchain=tool,
    )

    for src, dst in files:
        try:
            _resolve_expressions_in_file(src, dst, paths)
        except ExpressionError as err:
            _LOG.error('Error evaluating expressions in %s:', src)
            _LOG.error('  %s', err)
            return 1

    return 0


if __name__ == '__main__':
    sys.exit(main(**vars(_parse_args())))



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Wrapers for pyserial classes to log read and write data."""

import collections
import logging
from dataclasses import dataclass, field
import time
from typing import Optional

_LOG = logging.getLogger('pw_console')


@dataclass
class EventCountHistory:
    """Track counts of events over time.

    Example usage: ::

        events = EventCountHistory(
            base_count_units='Bytes',
            display_unit_title='KiB/s',
            display_unit_factor=0.001,
            interval=1.0,
            show_sparkline=True)

        # Log 1 event now.
        events.log(1)
        time.sleep(1)

        # Log 100 events at this time.
        events.log(100)
        time.sleep(1)

        events.log(200)
        time.sleep(1)
        events.log(400)
        print(events)
        ▂▄█ 0.400 [KiB/s]

    """

    base_count_units: str = 'Bytes'
    display_unit_title: str = 'KiB/s'
    display_unit_factor: float = 0.001
    interval: float = 1.0  # Number of seconds per sum of events.
    history_limit: int = 20
    scale_characters = ' ▁▂▃▄▅▆▇█'
    history: collections.deque = field(default_factory=collections.deque)
    show_sparkline: bool = False
    _this_count: int = 0
    _last_count: int = 0
    _last_update_time: float = field(default_factory=time.time)

    def log(self, count: int) -> None:
        self._this_count += count

        this_time = time.time()
        if this_time - self._last_update_time >= self.interval:
            self._last_update_time = this_time
            self._last_count = self._this_count
            self._this_count = 0
            self.history.append(self._last_count)

            if len(self.history) > self.history_limit:
                self.history.popleft()

    def last_count(self) -> float:
        return self._last_count * self.display_unit_factor

    def last_count_raw(self) -> int:
        return self._last_count

    def last_count_with_units(self) -> str:
        return '{:.3f} [{}]'.format(
            self._last_count * self.display_unit_factor, self.display_unit_title
        )

    def __repr__(self) -> str:
        sparkline = ''
        if self.show_sparkline:
            sparkline = self.sparkline()
        return ' '.join([sparkline, self.last_count_with_units()])

    def __pt_formatted_text__(self):
        return [('', self.__repr__())]

    def sparkline(
        self, min_value: int = 0, max_value: Optional[int] = None
    ) -> str:
        msg = ''.rjust(self.history_limit)
        if len(self.history) == 0:
            return msg

        minimum = min_value
        maximum = max_value if max_value else max(self.history)
        max_minus_min = maximum - min_value
        if max_minus_min == 0:
            return msg

        msg = ''
        for i in self.history:
            # (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min
            index = int(
                (((1.0 * i) - minimum) / max_minus_min)
                * len(self.scale_characters)
            )
            if index >= len(self.scale_characters):
                index = len(self.scale_characters) - 1
            msg += self.scale_characters[index]
        return msg.rjust(self.history_limit)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Facilities for generating TUF target metadata."""

import enum
import hashlib
from typing import Dict, Iterable

from pw_software_update.tuf_pb2 import (
    CommonMetadata,
    Hash,
    HashFunction,
    TargetFile,
    TargetsMetadata,
)

HASH_FACTORIES = {
    HashFunction.SHA256: hashlib.sha256,
}
DEFAULT_HASHES = (HashFunction.SHA256,)
DEFAULT_SPEC_VERSION = "0.0.1"
DEFAULT_METADATA_VERSION = 0


class RoleType(enum.Enum):
    """Set of allowed TUF metadata role types."""

    ROOT = 'root'
    TARGETS = 'targets'


def gen_common_metadata(
    role: RoleType,
    spec_version: str = DEFAULT_SPEC_VERSION,
    version: int = DEFAULT_METADATA_VERSION,
) -> CommonMetadata:
    """Generates CommonMetadata."""
    return CommonMetadata(
        role=role.value, spec_version=spec_version, version=version
    )


def gen_target_file(
    file_name: str, file_contents: bytes, hash_funcs=DEFAULT_HASHES
) -> TargetFile:
    return TargetFile(
        file_name=file_name,
        length=len(file_contents),
        hashes=gen_hashes(file_contents, hash_funcs),
    )


def gen_targets_metadata(
    target_payloads: Dict[str, bytes],
    hash_funcs: Iterable['HashFunction.V'] = DEFAULT_HASHES,
    version: int = DEFAULT_METADATA_VERSION,
) -> TargetsMetadata:
    """Generates TargetsMetadata the given target payloads."""
    target_files = []
    for target_file_name, target_payload in target_payloads.items():
        new_target_file = gen_target_file(
            file_name=target_file_name,
            file_contents=target_payload,
            hash_funcs=hash_funcs,
        )

        target_files.append(new_target_file)

    common_metadata = gen_common_metadata(RoleType.TARGETS, version=version)
    return TargetsMetadata(
        common_metadata=common_metadata, target_files=target_files
    )


def gen_hashes(
    data: bytes, hash_funcs: Iterable['HashFunction.V']
) -> Iterable[Hash]:
    """Computes all the specified hashes over the data."""
    result = []
    for func in hash_funcs:
        if func == HashFunction.UNKNOWN_HASH_FUNCTION:
            raise ValueError(
                'UNKNOWN_HASH_FUNCTION cannot be used to generate hashes.'
            )
        digest = HASH_FACTORIES[func](data).digest()
        result.append(Hash(function=func, hash=digest))

    return result
